# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
from ..tools import axes_check
import atomman.unitconvert as uc

class ElasticConstants(object):
    """Class for storing and converting elastic constant values"""
    
    def __init__(self, **kwargs):
        """
        Initilizes an ElasticConstants instance from one of the parameter
        options.
        
        Parameters
        ----------
        Cij : numpy.ndarray
            (6, 6) array of Voigt representation of elastic stiffness.
        Sij : numpy.ndarray
            (6, 6) array of Voigt representation of elastic compliance.
        Cij9 : numpy.ndarray
            (9, 9) array representation of elastic stiffness.
        Cijkl : numpy.ndarray
            (3, 3, 3, 3) array representation of elastic stiffness.
        Sijkl : numpy.ndarray
            (3, 3, 3, 3) array representation of elastic compliance.
        model : DataModelDict, string, or file-like object 
            Data model containing elastic constants.
        C11, C12, ... C66 : float
            Individual components of Cij for a standardized representation:
            isotropic: C11, C12 (see below for more options)
            cubic: C11, C12, C44
            hexagonal: C11, C12, C13, C33, C44, C66 (2*C66=C11-C12)
            tetragonal: C11, C12, C13, C16, C33, C44, C66 (C16 optional)
            rhombohedral: C11, C12, C13, C14, C15, C33, C44, C66 (2*C66=C11-C12, C15 optional)
            orthorhombic: C11, C12, C13, C22, C23, C33, C44, C55, C66
            monoclinic: C11, C12, C13, C15, C22, C23, C25, C33, C35, C44, C46, C55, C66
            triclinic: all Cij where i <= j
        M, lambda, mu, E, nu, K, C11, C12, C44 : float
            Individual isotropic elastic moduli (exactly two must be given).
        """
        # Initialize for no arguments
        if len(kwargs) == 0:
            self.__c_ij = np.zeros((6,6), dtype='float64')
        
        # Initialize for matrix arguments
        elif 'Cij' in kwargs:
            assert len(kwargs) == 1, 'Cij cannot be specified with other keyword arguments'
            self.Cij = kwargs['Cij']
        elif 'Sij' in kwargs:
            assert len(kwargs) == 1, 'Sij cannot be specified with other keyword arguments'
            self.Sij = kwargs['Sij']
        elif 'Cij9' in kwargs:
            assert len(kwargs) == 1, 'Cij9 cannot be specified with other keyword arguments'
            self.Cij9 = kwargs['Cij9']
        elif 'Cijkl' in kwargs:
            assert len(kwargs) == 1, 'Cijkl cannot be specified with other keyword arguments'
            self.Cijkl = kwargs['Cijkl']
        elif 'Sijkl' in kwargs:
            assert len(kwargs) == 1, 'Sijkl cannot be specified with other keyword arguments'
            self.Sijkl = kwargs['Sijkl']
        
        # Initialize using data model
        elif 'model' in kwargs:
            self.model(**kwargs)
        
        # Initialize for individually specified parameters
        elif len(kwargs) == 2:
            self.isotropic(**kwargs)
        elif len(kwargs) == 3:
            self.cubic(**kwargs)
        elif len(kwargs) == 5:
            self.hexagonal(**kwargs)
        elif len(kwargs) == 6 or len(kwargs) == 7:
            if 'C14' in kwargs:
                self.rhombohedral(**kwargs)
            else:
                self.tetragonal(**kwargs)
        elif len(kwargs) == 8:
            self.rhombohedral(**kwargs)
        elif len(kwargs) == 9:
            self.orthorhombic(**kwargs)
        elif len(kwargs) == 13:
            self.monoclinic(**kwargs)
        elif len(kwargs) == 21:
            self.triclinic(**kwargs)
        else:
            raise TypeError('Invalid argument keywords')
    
    def __str__(self):
        """Calling string returns str(self.Cij)."""
        return str(self.Cij)
    
    @property
    def Cij(self):
        """The stiffness constants in Voigt 6x6 format"""
        return deepcopy(self.__c_ij)
    
    @Cij.setter
    def Cij(self, value):
        value = np.asarray(value, dtype='float64')
        assert value.shape == (6,6),  'Cij must be 6x6'
        
        # Zero out near-zero terms
        assert value.max() > 0.0, 'Cij values not valid'
        value[np.isclose(value/value.max(), 0.0, atol=1e-9)] = 0.0
        
        # Check symmetry
        for i in range(6):
            for j in range(i):
                assert np.isclose(value[i,j], value[j,i], atol=1e-9), '6x6 matrix not symmetric!' 
        self.__c_ij = value
    
    @property
    def Sij(self):
        """The compliance constants in Voigt 6x6 format"""
        return np.linalg.inv(self.Cij)
    
    @Sij.setter
    def Sij(self, value):
        value = np.asarray(value, dtype='float64')
        assert value.shape == (6,6),  'Sij must be 6x6'
        self.Cij = np.linalg.inv(value)
    
    @property
    def Cij9(self):
        """The stiffness constants in 9x9 format"""
        c = self.Cij
        return np.array([[c[0,0],c[0,1],c[0,2],c[0,3],c[0,4],c[0,5],c[0,3],c[0,4],c[0,5]],
                         [c[1,0],c[1,1],c[1,2],c[1,3],c[1,4],c[1,5],c[1,3],c[1,4],c[1,5]],
                         [c[2,0],c[2,1],c[2,2],c[2,3],c[2,4],c[2,5],c[2,3],c[2,4],c[2,5]],
                         [c[3,0],c[3,1],c[3,2],c[3,3],c[3,4],c[3,5],c[3,3],c[3,4],c[3,5]],
                         [c[4,0],c[4,1],c[4,2],c[4,3],c[4,4],c[4,5],c[4,3],c[4,4],c[4,5]],
                         [c[5,0],c[5,1],c[5,2],c[5,3],c[5,4],c[5,5],c[5,3],c[5,4],c[5,5]],
                         [c[3,0],c[3,1],c[3,2],c[3,3],c[3,4],c[3,5],c[3,3],c[3,4],c[3,5]],
                         [c[4,0],c[4,1],c[4,2],c[4,3],c[4,4],c[4,5],c[4,3],c[4,4],c[4,5]],
                         [c[5,0],c[5,1],c[5,2],c[5,3],c[5,4],c[5,5],c[5,3],c[5,4],c[5,5]]])
    
    @Cij9.setter
    def Cij9(self, value):
        value = np.asarray(value, dtype='float64')
        assert value.shape == (9,9), 'Cij9 must be 9x9'
        
        # Check symmetry
        for i in range(6, 9):
            for j in range(9):
                assert value[i,j] == value[i-3, j]
                assert value[j,i] == value[j, i-3]
        self.Cij = value[:6, :6]
    
    @property
    def Cijkl(self):
        """The stiffness constants in 3x3x3x3 format"""
        c = self.Cij
        return np.array([[[[c[0,0],c[0,5],c[0,4]], [c[0,5],c[0,1],c[0,3]], [c[0,4],c[0,3],c[0,2]]],
                          [[c[5,0],c[5,5],c[5,4]], [c[5,5],c[5,1],c[5,3]], [c[5,4],c[5,3],c[5,2]]],
                          [[c[4,0],c[4,5],c[4,4]], [c[4,5],c[4,1],c[4,3]], [c[4,4],c[4,3],c[4,2]]]], 
                        
                         [[[c[5,0],c[5,5],c[5,4]], [c[5,5],c[5,1],c[5,3]], [c[5,4],c[5,3],c[5,2]]],
                          [[c[1,0],c[1,5],c[1,4]], [c[1,5],c[1,1],c[1,3]], [c[1,4],c[1,3],c[1,2]]],
                          [[c[3,0],c[3,5],c[3,4]], [c[3,5],c[3,1],c[3,3]], [c[3,4],c[3,3],c[3,2]]]],
                         
                         [[[c[4,0],c[4,5],c[4,4]], [c[4,5],c[4,1],c[4,3]], [c[4,4],c[4,3],c[4,2]]],
                          [[c[3,0],c[3,5],c[3,4]], [c[3,5],c[3,1],c[3,3]], [c[3,4],c[3,3],c[3,2]]],
                          [[c[2,0],c[2,5],c[2,4]], [c[2,5],c[2,1],c[2,3]], [c[2,4],c[2,3],c[2,2]]]]])
    
    @Cijkl.setter
    def Cijkl(self, value):
        c = np.asarray(value, dtype='float64')
        assert c.shape == (3,3,3,3),  'Cijkl must be 3x3x3x3'
        assert c.max() > 0.0, 'Cij values not valid'
        # Check symmetry
        indexes = np.array([[0,0], [1,1], [2,2], [1,2], [0,2], [0,1]], dtype=int)
        for ij in range(6):
            for kl in range(ij, 6):
                i, j, k, l = indexes[ij,0], indexes[ij,1], indexes[kl,0], indexes[kl,1]
                assert np.isclose(c[i,j,k,l], c[j,i,k,l])
                assert np.isclose(c[i,j,k,l], c[j,i,l,k])
                assert np.isclose(c[i,j,k,l], c[k,l,j,i])
                assert np.isclose(c[i,j,k,l], c[l,k,j,i])
                assert np.isclose(c[i,j,k,l], c[i,j,l,k])
                assert np.isclose(c[i,j,k,l], c[k,l,i,j])
                assert np.isclose(c[i,j,k,l], c[l,k,i,j])
        
        self.Cij = np.array([[c[0,0,0,0], c[0,0,1,1], c[0,0,2,2], c[0,0,1,2], c[0,0,0,2], c[0,0,0,1]],
                             [c[1,1,0,0], c[1,1,1,1], c[1,1,2,2], c[1,1,1,2], c[1,1,0,2], c[1,1,0,1]],
                             [c[2,2,0,0], c[2,2,1,1], c[2,2,2,2], c[2,2,1,2], c[2,2,0,2], c[2,2,0,1]],
                             [c[1,2,0,0], c[1,2,1,1], c[1,2,2,2], c[1,2,1,2], c[1,2,0,2], c[1,2,0,1]],
                             [c[0,2,0,0], c[0,2,1,1], c[0,2,2,2], c[0,2,1,2], c[0,2,0,2], c[0,2,0,1]],
                             [c[0,1,0,0], c[0,1,1,1], c[0,1,2,2], c[0,1,1,2], c[0,1,0,2], c[0,1,0,1]]])
    
    @property
    def Sijkl(self):
        """The compliance constants in 3x3x3x3 format"""
        s = self.Sij
        s[3:,:] = s[3:,:]/2.
        s[:,3:] = s[:,3:]/2.
        return np.array([[[[s[0,0],s[0,5],s[0,4]], [s[0,5],s[0,1],s[0,3]], [s[0,4],s[0,3],s[0,2]]],
                          [[s[5,0],s[5,5],s[5,4]], [s[5,5],s[5,1],s[5,3]], [s[5,4],s[5,3],s[5,2]]],
                          [[s[4,0],s[4,5],s[4,4]], [s[4,5],s[4,1],s[4,3]], [s[4,4],s[4,3],s[4,2]]]],
                         
                         [[[s[5,0],s[5,5],s[5,4]], [s[5,5],s[5,1],s[5,3]], [s[5,4],s[5,3],s[5,2]]],
                          [[s[1,0],s[1,5],s[1,4]], [s[1,5],s[1,1],s[1,3]], [s[1,4],s[1,3],s[1,2]]],
                          [[s[3,0],s[3,5],s[3,4]], [s[3,5],s[3,1],s[3,3]], [s[3,4],s[3,3],s[3,2]]]],
                         
                         [[[s[4,0],s[4,5],s[4,4]], [s[4,5],s[4,1],s[4,3]], [s[4,4],s[4,3],s[4,2]]],
                          [[s[3,0],s[3,5],s[3,4]], [s[3,5],s[3,1],s[3,3]], [s[3,4],s[3,3],s[3,2]]],
                          [[s[2,0],s[2,5],s[2,4]], [s[2,5],s[2,1],s[2,3]], [s[2,4],s[2,3],s[2,2]]]]])
    
    @Sijkl.setter
    def Sijkl(self, value):
        s = np.asarray(value, dtype='float64')
        assert s.shape == (3,3,3,3),  'Sijkl must be 3x3x3x3'
        
        # Check symmetry
        indexes = np.array([[0,0], [1,1], [2,2], [1,2], [0,2], [0,1]], dtype=int)
        for ij in range(6):
            for kl in range(ij, 6):
                i, j, k, l = indexes[ij,0], indexes[ij,1], indexes[kl,0], indexes[kl,1]
                assert np.isclose(s[i,j,k,l], s[j,i,k,l])
                assert np.isclose(s[i,j,k,l], s[j,i,l,k])
                assert np.isclose(s[i,j,k,l], s[k,l,j,i])
                assert np.isclose(s[i,j,k,l], s[l,k,j,i])
                assert np.isclose(s[i,j,k,l], s[i,j,l,k])
                assert np.isclose(s[i,j,k,l], s[k,l,i,j])
                assert np.isclose(s[i,j,k,l], s[l,k,i,j]) 
        
        self.Sij = np.array([[   s[0,0,0,0],    s[0,0,1,1],    s[0,0,2,2], 2.*s[0,0,1,2], 2.*s[0,0,0,2], 2.*s[0,0,0,1]],
                             [   s[1,1,0,0],    s[1,1,1,1],    s[1,1,2,2], 2.*s[1,1,1,2], 2.*s[1,1,0,2], 2.*s[1,1,0,1]],
                             [   s[2,2,0,0],    s[2,2,1,1],    s[2,2,2,2], 2.*s[2,2,1,2], 2.*s[2,2,0,2], 2.*s[2,2,0,1]],
                             [2.*s[1,2,0,0], 2.*s[1,2,1,1], 2.*s[1,2,2,2], 4.*s[1,2,1,2], 4.*s[1,2,0,2], 4.*s[1,2,0,1]],
                             [2.*s[0,2,0,0], 2.*s[0,2,1,1], 2.*s[0,2,2,2], 4.*s[0,2,1,2], 4.*s[0,2,0,2], 4.*s[0,2,0,1]],
                             [2.*s[0,1,0,0], 2.*s[0,1,1,1], 2.*s[0,1,2,2], 4.*s[0,1,1,2], 4.*s[0,1,0,2], 4.*s[0,1,0,1]]])
    
    def transform(self, axes, tol=1e-8):
        """
        Transforms the elastic constant matrix based on the supplied axes.
        
        Parameters
        ----------
        axes : numpy.ndarray
            (3, 3) array giving three right-handed orthogonal vectors to use
            for transforming.
        tol : float, optional
            Relative tolerance to use in identifying near-zero terms.
            
        Returns
        -------
        ElasticConstants
            A new ElasticConstants object that has been transformed.
        """
        axes = np.asarray(axes, dtype='float64')
        T = axes_check(axes)
        
        Q = np.einsum('km,ln->mnkl', T, T)
        C = np.einsum('ghij,ghmn,mnkl->ijkl', Q, self.Cijkl, Q)
        C[abs(C / C.max()) < tol] = 0.0
        
        return ElasticConstants(Cijkl=C)
    
    def isotropic(self, **kwargs):
        """
        Set values with two independent isotropic moduli.
        
        Parameters
        ----------
        C11 : float, optional
            C11 component of Cij.
        C12 : float, optional
            C12 component of Cij.
        C44 : float, optional
            C44 component of Cij.
        M : float, optional
            P-wave modulus(Equivalent to C11).
        lambda : float, optional
            Lame's first parameter (Equivalent to C12).
        mu : float, optional
            Shear modulus (Equivalent to C44).
        E : float, optional
            Young's modulus
        nu : float, optional
            Poisson's ratio
        K : float, optional
            Bulk modulus
        """
        
        try:
            # Handle equivalent terms
            if 'M' in kwargs:
                kwargs['C11'] = kwargs.pop('M')
            if 'lambda' in kwargs:
                kwargs['C12'] = kwargs.pop('lambda')
            if 'mu' in kwargs:
                kwargs['C44'] = kwargs.pop('mu')
            
            # Check len of kwargs
            assert len(kwargs) == 2
            
            # Pop and convert terms
            if 'C11' in kwargs:
                c11 = kwargs.pop('C11')
                
                if 'C12' in kwargs:
                    c12 = kwargs.pop('C12')
                    c44 = (c11 - c12) / 2
                
                else:
                    if 'C44' in kwargs:
                        c44 = kwargs.pop('C44')
                    
                    elif 'E' in kwargs: 
                        E = kwargs.pop('E')
                        S = (E**2 + 9 * c11**2 - 10 * E * c11)**0.5
                        c44 = (3 * c11 + E - S) / 8
                    
                    elif 'nu' in kwargs: 
                        nu = kwargs.pop('nu')
                        c44 = c11 * (1 - 2 * nu) / (2 * (1 - nu))
                    
                    elif 'K' in kwargs: 
                        K = kwargs.pop('K')
                        c44 = 3 * (c11 - K) / 4
                    
                    c12 = c11 - 2 * c44
            
            else:
                if 'C12' in kwargs:
                    c12 = kwargs.pop('C12')
                    
                    if 'C44' in kwargs:
                        c44 = kwargs.pop('C44')
                    
                    elif 'E' in kwargs:
                        E = kwargs.pop('E')
                        R = (E**2 + 9 * c12**2 + 2 * E * c12)**0.5
                        c44 = (E - 3 * c12 + R) / 4
                    
                    elif 'nu' in kwargs:
                        nu = kwargs.pop('nu')
                        c44 = c12 * (1 - 2 * nu) / (2 * nu)
                    
                    elif 'K' in kwargs:
                        K = kwargs.pop('K')
                        c44 = 3 * (K - c12) / 2
                
                elif 'C44' in kwargs:
                    c44 = kwargs.pop('C44')
                    
                    if 'E' in kwargs:
                        E = kwargs.pop('E')
                        c12 = c44 * (E - 2 * c44) / (3 * c44 - E)
                    
                    elif 'nu' in kwargs:
                        nu = kwargs.pop('nu')
                        c12 = 2 * c44 * nu / (1 - 2 * nu)
                    
                    elif 'K' in kwargs:
                        K = kwargs.pop('K')
                        c12 = K - 2 * c44 / 3
                
                elif 'E' in kwargs: 
                    E = kwargs.pop('E')
                    
                    if 'nu' in kwargs: 
                        nu = kwargs.pop('nu')
                        c12 = E * nu / ((1 + nu) * (1 - 2 * nu))
                        c44 = E / (2 * (1 + nu))
                    
                    elif 'K' in kwargs: 
                        K = kwargs.pop('K')
                        c12 = 3 * K * (3 * K - E) / (9 * K - E)
                        c44 = 3 * K * E / (9 * K - E)
                
                elif 'nu' in kwargs:
                    nu = kwargs.pop('nu')
                    
                    if 'K' in kwargs: 
                        K = kwargs.pop('K')
                        c12 = 3 * K * nu / (1 + nu)
                        c44 = 3 * K * (1 - 2 * nu) / (2 * (1 + nu))
                
                c11 = c12 + 2 * c44
        
        except:
            raise TypeError('isotropic style takes two unique keyword arguments of (C11=M, C12=lambda, C44=mu, E, nu, K)')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c12, 0.0, 0.0, 0.0],
                             [c12, c11, c12, 0.0, 0.0, 0.0],
                             [c12, c12, c11, 0.0, 0.0, 0.0],
                             [0.0, 0.0, 0.0, c44, 0.0, 0.0],
                             [0.0, 0.0, 0.0, 0.0, c44, 0.0],
                             [0.0, 0.0, 0.0, 0.0, 0.0, c44]]) 
    
    def cubic(self, **kwargs):
        """
        Set values with three independent cubic moduli.
        
        Parameters
        ----------
        C11 : float
            C11 component of Cij.
        C12 : float
            C12 component of Cij.
        C44 : float
            C44 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) == 3
            
            # Pop required independent terms
            c11 = kwargs.pop('C11')
            c12 = kwargs.pop('C12')
            c44 = kwargs.pop('C44')
        except:
            raise TypeError('cubic style takes keyword arguments C11, C12, and C66')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c12, 0.0, 0.0, 0.0],
                             [c12, c11, c12, 0.0, 0.0, 0.0],
                             [c12, c12, c11, 0.0, 0.0, 0.0],
                             [0.0, 0.0, 0.0, c44, 0.0, 0.0],
                             [0.0, 0.0, 0.0, 0.0, c44, 0.0],
                             [0.0, 0.0, 0.0, 0.0, 0.0, c44]])
    
    def hexagonal(self, **kwargs):
        """
        Set values with five independent hexagonal moduli.
        (2 * C66 = C11 - C12)
        
        Parameters
        ----------
        C11 : float, optional
            C11 component of Cij.
        C12 : float, optional
            C12 component of Cij.
        C13 : float
            C13 component of Cij.
        C33 : float
            C33 component of Cij.
        C44 : float
            C44 component of Cij.
        C66 : float, optional
            C66 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) >= 5 and len(kwargs) <=6
            
            # Pop required independent terms
            c33 = kwargs.pop('C33')
            c13 = kwargs.pop('C13')
            c44 = kwargs.pop('C44')
            
            # Pop required dependent terms
            if 'C11' in kwargs and 'C12' in kwargs:
                c11 = kwargs.pop('C11')
                c12 = kwargs.pop('C12')
                c66 = (c11 - c12) / 2
                
                # Check if redundant C66 is given
                if 'C66' in kwargs:
                    assert np.isclose(c66, kwargs['C66'])
                    c66 = kwargs.pop('C66')
            
            elif 'C11' in kwargs and 'C66' in kwargs:
                c11 = kwargs.pop('C11')
                c66 = kwargs.pop('C66')
                c12 = c11 - 2 * c66
            
            elif 'C12' in kwargs and 'C66' in kwargs:
                c12 = kwargs.pop('C12')
                c66 = kwargs.pop('C66')
                c11 = 2 * c66 + c12
            else:
                assert False
        except:
            raise TypeError('hexagonal style takes keyword arguments C33, C13, C44, and two of (2*C66=C11-C12)')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c13, 0.0, 0.0, 0.0],
                             [c12, c11, c13, 0.0, 0.0, 0.0],
                             [c13, c13, c33, 0.0, 0.0, 0.0],
                             [0.0, 0.0, 0.0, c44, 0.0, 0.0],
                             [0.0, 0.0, 0.0, 0.0, c44, 0.0],
                             [0.0, 0.0, 0.0, 0.0, 0.0, c66]])
    
    def rhombohedral(self, **kwargs):
        """
        Set values with six or seven independent rhombohedral moduli.
        (2 * C66 = C11 - C12)
        
        Parameters
        ----------
        C11 : float, optional
            C11 component of Cij.
        C12 : float, optional
            C12 component of Cij.
        C13 : float
            C13 component of Cij.
        C14 : float
            C33 component of Cij.
        C15 : float, optional
            C44 component of Cij.
        C33 : float
            C66 component of Cij.
        C44 : float
            C66 component of Cij.
        C66 : float, optional
            C66 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) >= 6 and len(kwargs) <= 8
            
            # Pop required independent terms
            c33 = kwargs.pop('C33')
            c13 = kwargs.pop('C13')
            c14 = kwargs.pop('C14')
            c44 = kwargs.pop('C44')
            
            # Pop required dependent terms
            if 'C11' in kwargs and 'C12' in kwargs:
                c11 = kwargs.pop('C11')
                c12 = kwargs.pop('C12')
                c66 = (c11 - c12) / 2
                
                # Check if redundant C66 is given
                if 'C66' in kwargs:
                    assert np.isclose(c66, kwargs['C66'])
                    c66 = kwargs.pop('C66')
            
            elif 'C11' in kwargs and 'C66' in kwargs:
                c11 = kwargs.pop('C11')
                c66 = kwargs.pop('C66')
                c12 = c11 - 2 * c66
            
            elif 'C12' in kwargs and 'C66' in kwargs:
                c12 = kwargs.pop('C12')
                c66 = kwargs.pop('C66')
                c11 = 2 * c66 + c12
            else:
                assert False
            
            # Check for optional term
            if len(kwargs) == 0:
                c15 = 0.0
            else:
                c15 = kwargs.pop('C15')
                assert len(kwargs) == 0
        except:
            raise TypeError('rhombohedral style takes keyword arguments C33, C13, C14, C44, at least two of (2*C66=C11-C12) and optional C15')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c13, c14, c15, 0.0],
                             [c12, c11, c13,-c14,-c15, 0.0],
                             [c13, c13, c33, 0.0, 0.0, 0.0],
                             [c14,-c14, 0.0, c44, 0.0,-c15],
                             [c15,-c15, 0.0, 0.0, c44, c14],
                             [0.0, 0.0, 0.0,-c15, c14, c66]])
    
    def tetragonal(self, **kwargs):
        """
        Set values with six or seven independent tetragonal moduli.
        
        Parameters
        ----------
        C11 : float
            C11 component of Cij.
        C12 : float
            C12 component of Cij.
        C13 : float
            C13 component of Cij.
        C16 : float, optional
            C16 component of Cij.
        C33 : float
            C33 component of Cij.
        C44 : float
            C44 component of Cij.
        C66 : float
            C66 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) == 6 or len(kwargs) == 7
            
            # Pop required independent terms
            c11 = kwargs.pop('C11')
            c33 = kwargs.pop('C33')
            c12 = kwargs.pop('C12')
            c13 = kwargs.pop('C13')
            c44 = kwargs.pop('C44')
            c66 = kwargs.pop('C66')
            
            # Check for optional term
            if len(kwargs) == 0:
                c16 = 0.0
            else:
                c16 = kwargs.pop('C16')
                assert len(kwargs) == 0
        except:
            raise TypeError('tetragonal style takes keyword arguments C11, C33, C12, C13, C44, C66, and optional C16')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c13, 0.0, 0.0, c16],
                             [c12, c11, c13, 0.0, 0.0,-c16],
                             [c13, c13, c33, 0.0, 0.0, 0.0],
                             [0.0, 0.0, 0.0, c44, 0.0, 0.0],
                             [0.0, 0.0, 0.0, 0.0, c44, 0.0],
                             [c16,-c16, 0.0, 0.0, 0.0, c66]])
    
    def orthorhombic(self, **kwargs):
        """
        Set values with nine independent orthorhombic moduli.
        
        Parameters
        ----------
        C11 : float
            C11 component of Cij.
        C12 : float
            C12 component of Cij.
        C13 : float
            C13 component of Cij.
        C22 : float
            C22 component of Cij.
        C23 : float
            C23 component of Cij.
        C33 : float
            C33 component of Cij.
        C44 : float
            C44 component of Cij.
        C55 : float
            C55 component of Cij.
        C66 : float
            C66 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) == 9
            
            # Set required independent terms
            c11 = kwargs['C11']
            c22 = kwargs['C22']
            c33 = kwargs['C33']
            c12 = kwargs['C12']
            c13 = kwargs['C13']
            c23 = kwargs['C23']
            c44 = kwargs['C44']
            c55 = kwargs['C55']
            c66 = kwargs['C66']
        except:
            raise TypeError('orthorhombic style takes keyword arguments C11, C22, C33, C12, C13, C23, C44, C55, C66')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c13, 0.0, 0.0, 0.0],
                             [c12, c22, c23, 0.0, 0.0, 0.0],
                             [c13, c23, c33, 0.0, 0.0, 0.0],
                             [0.0, 0.0, 0.0, c44, 0.0, 0.0],
                             [0.0, 0.0, 0.0, 0.0, c55, 0.0],
                             [0.0, 0.0, 0.0, 0.0, 0.0, c66]])
    
    def monoclinic(self, **kwargs):
        """
        Set values with thirteen independent monoclinic moduli.
        
        Parameters
        ----------
        C11 : float
            C11 component of Cij.
        C12 : float
            C12 component of Cij.
        C13 : float
            C13 component of Cij.
        C15 : float
            C15 component of Cij.
        C22 : float
            C22 component of Cij.
        C23 : float
            C23 component of Cij.
        C25 : float
            C25 component of Cij.
        C33 : float
            C33 component of Cij.
        C35 : float
            C35 component of Cij.
        C44 : float
            C44 component of Cij.
        C46 : float
            C46 component of Cij.
        C55 : float
            C55 component of Cij.
        C66 : float
            C66 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) == 13
            
            # Set required independent terms
            c11 = kwargs['C11']
            c12 = kwargs['C12']
            c13 = kwargs['C13']
            c15 = kwargs['C15']
            c22 = kwargs['C22']
            c23 = kwargs['C23']
            c25 = kwargs['C25']
            c33 = kwargs['C33']
            c35 = kwargs['C35']
            c44 = kwargs['C44']
            c46 = kwargs['C46']
            c55 = kwargs['C55']
            c66 = kwargs['C66']
        except:
            raise TypeError('monoclinic style takes keyword arguments C11, C12, C13, C15, C22, C23, C25, C33, C35, C44, C46, C55, C66')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c13, 0.0, c15, 0.0],
                             [c12, c22, c23, 0.0, c25, 0.0],
                             [c13, c23, c33, 0.0, c35, 0.0],
                             [0.0, 0.0, 0.0, c44, 0.0, c46],
                             [c15, c25, c35, 0.0, c55, 0.0],
                             [0.0, 0.0, 0.0, c46, 0.0, c66]])
    
    def triclinic(self, **kwargs):
        """
        Set values with twenty one independent triclinic moduli
        
        Parameters
        ----------
        C11 : float
            C11 component of Cij.
        C12 : float
            C12 component of Cij.
        C13 : float
            C13 component of Cij.
        C14 : float
            C14 component of Cij.
        C15 : float
            C15 component of Cij.
        C16 : float
            C16 component of Cij.
        C22 : float
            C22 component of Cij.
        C23 : float
            C23 component of Cij.
        C24 : float
            C24 component of Cij.
        C25 : float
            C25 component of Cij.
        C26 : float
            C26 component of Cij.
        C33 : float
            C33 component of Cij.
        C34 : float
            C34 component of Cij.
        C35 : float
            C35 component of Cij.
        C36 : float
            C36 component of Cij.
        C44 : float
            C44 component of Cij.
        C45 : float
            C45 component of Cij.
        C46 : float
            C46 component of Cij.
        C55 : float
            C55 component of Cij.
        C56 : float
            C56 component of Cij.
        C66 : float
            C66 component of Cij.
        """
        
        try:
            # Check len of kwargs
            assert len(kwargs) == 21
            
            # Set required independent terms
            c11 = kwargs['C11']
            c12 = kwargs['C12']
            c13 = kwargs['C13']
            c14 = kwargs['C14']
            c15 = kwargs['C15']
            c16 = kwargs['C16']
            c22 = kwargs['C22']
            c23 = kwargs['C23']
            c24 = kwargs['C24']
            c25 = kwargs['C25']
            c26 = kwargs['C26']
            c33 = kwargs['C33']
            c34 = kwargs['C34']
            c35 = kwargs['C35']
            c36 = kwargs['C36']
            c44 = kwargs['C44']
            c45 = kwargs['C45']
            c46 = kwargs['C46']
            c55 = kwargs['C55']
            c56 = kwargs['C56']
            c66 = kwargs['C66']
        except:
            raise TypeError('triclinic style takes keyword arguments of all Cij where i <= j')
        
        # Build Cij array
        self.Cij = np.array([[c11, c12, c13, c14, c15, c16],
                             [c12, c22, c23, c24, c25, c26],
                             [c13, c23, c33, c34, c35, c36],
                             [c14, c24, c34, c44, c45, c46],
                             [c15, c25, c35, c45, c55, c56],
                             [c16, c26, c36, c46, c56, c66]])
    
    def normalized_as(self, crystal_system):
        """
        Returns a new ElasticConstants object where values of the current are
        averaged or zeroed out according to a standard crystal system setting.
        NOTE: no validation checks are made to evaluate whether such
        normalizations should be done! That is left up to you (compare values
        before and after normalization).
        
        Parameters
        ----------
        crystal_system : str
            Indicates the crystal system representation to use when building a
            data model.
            
        Returns
        -------
        atomman.ElasticConstants
            The elastic constants normalized according to the crystal system
            symmetries.
        """
        
        if crystal_system == 'triclinic':
            return ElasticConstants(Cij = self.Cij)
        
        else:
            c = self.Cij
            c_dict = {}
            
            if crystal_system == 'isotropic':
                c_dict['mu'] = self.shear()
                c_dict['K'] = self.bulk()
            
            elif crystal_system == 'cubic':
                c_dict['C11'] = (c[0,0] + c[1,1] + c[2,2]) / 3
                c_dict['C12'] = (c[0,1] + c[0,2] + c[1,2]) / 3
                c_dict['C44'] = (c[3,3] + c[4,4] + c[5,5]) / 3
            
            elif crystal_system == 'hexagonal':
                c_dict['C11'] = (c[0,0] + c[1,1]) / 2
                c_dict['C33'] = c[2,2]
                c_dict['C12'] = (c[0,1] + (c[0,0] - 2*c[5,5])) / 2
                c_dict['C13'] = (c[0,2] + c[1,2]) / 2
                c_dict['C44'] = (c[3,3] + c[4,4]) / 2
            
            elif crystal_system == 'tetragonal':
                c_dict['C11'] = (c[0,0] + c[1,1]) / 2
                c_dict['C33'] = c[2,2]
                c_dict['C12'] = c[0,1]
                c_dict['C13'] = (c[0,2] + c[1,2]) / 2
                c_dict['C16'] = (c[0,5] - c[1,5]) / 2
                c_dict['C44'] = (c[3,3] + c[4,4]) / 2
                c_dict['C66'] = c[5,5]
            
            elif crystal_system == 'rhombohedral':
                c_dict['C11'] = (c[0,0] + c[1,1]) / 2
                c_dict['C33'] = c[2,2]
                c_dict['C12'] = (c[0,1] + (c[0,0] - 2*c[5,5])) / 2
                c_dict['C13'] = (c[0,2] + c[1,2]) / 2
                c_dict['C14'] = (c[0,3] - c[1,3]) / 2
                c_dict['C15'] = (c[0,4] - c[1,4] - c[3,5]) / 3
                c_dict['C44'] = (c[3,3] + c[4,4]) / 2
            
            elif crystal_system == 'orthorhombic':
                c_dict['C11'] = c[0,0]
                c_dict['C22'] = c[1,1]
                c_dict['C33'] = c[2,2]
                c_dict['C12'] = c[0,1]
                c_dict['C13'] = c[0,2]
                c_dict['C23'] = c[1,2]
                c_dict['C44'] = c[3,3]
                c_dict['C55'] = c[4,4]
                c_dict['C66'] = c[5,5]
            else:
                raise ValueError('Invalid crystal_system: ' + crystal_system)
            
            return ElasticConstants(**c_dict)
    
    def is_normal(self, crystal_system, atol=1e-4, rtol=1e-4):
        """
        Checks if current elastic constants agree with values normalized to
        a specified crystal family (within tolerances).
        
        Parameters
        ----------
        crystal_system : str
            Indicates the crystal system representation to use when building a
            data model.
        atol : float, optional
            Absolute tolerance to use.  Default value is 1e-4.
        rtol : float, optional
            Relative tolerance to use.  Default value is 1e-4.
        
        Returns
        -------
        bool
            True if all Cij match within the tolerances, false otherwise.
        """
        return np.allclose(self.Cij, self.normalized_as(crystal_system).Cij,
                           atol=atol, rtol=rtol)
    
    def model(self, model=None, unit=None, crystal_system='triclinic'):
        """
        Return or set DataModelDict representation of the elastic constants.
        
        Parameters
        ----------
        model : DataModelDict, string, or file-like object, optional
            Data model containing exactly one 'elastic-constants' branch to
            read.
        unit : str, optional
            Units or pressure to save values in when building a data model.
            Default value is None (no conversion).
        crystal_system : str, optional
            Indicates the crystal system representation to normalize by.
            Default value is 'triclinic', i.e. no normalization.
        
        Returns
        -------
        DataModelDict
            If model is not given as a parameter.
        """
        
        # Set values if model given
        if model is not None:
            
            # Find elastic-constants element
            model = DM(model).find('elastic-constants')
            
            # Read in values
            try:
                # New format
                self.Cij = uc.value_unit(model['Cij'])
            except:
                # Old format
                c_dict = {}
                for C in model['C']:
                    key = 'C' + C['ij'][0] + C['ij'][2]
                    c_dict[key] = uc.value_unit(C['stiffness'])
                self.Cij = ElasticConstants(**c_dict).Cij 
        
        # Return DataModelDict if model not given
        else:
            normCij = self.normalized_as(crystal_system).Cij
            model = DM()
            model['elastic-constants'] = DM()
            model['elastic-constants']['Cij'] = uc.model(normCij, unit)
            
            return model
    
    def bulk(self, style='Hill'):
        """
        Returns a bulk modulus estimate.
        
        Parameters
        ----------
        style : str
            Indicates which style of estimate to use.  Default value is 'Hill'.
            - 'Hill' -- Hill estimate (average of Voigt and Reuss).
            - 'Voigt' -- Voigt estimate. Uses Cij.
            - 'Reuss' -- Reuss estimate. Uses Sij.
        """
        if style == 'Hill':
            return (self.bulk('Voigt') + self.bulk('Reuss')) / 2
        
        elif style == 'Voigt':
            c = self.Cij
            return ( (c[0,0] + c[1,1] + c[2,2]) + 2*(c[0,1] + c[1,2] + c[0,2]) ) / 9
        
        elif style == 'Reuss':
            s = self.Sij
            return 1 / ( (s[0,0] + s[1,1] + s[2,2]) + 2*(s[0,1] + s[1,2] + s[0,2]) )
        
        else:
            raise ValueError('Unknown estimate style')
    
    def shear(self, style='Hill'):
        """
        Returns a shear modulus estimate.
        
        Parameters
        ----------
        style : str
            Indicates which style of estimate to use.  Default value is 'Hill'.
            - 'Hill' -- Hill estimate (average of Voigt and Reuss).
            - 'Voigt' -- Voigt estimate. Uses Cij.
            - 'Reuss' -- Reuss estimate. Uses Sij.
        """
        if style == 'Hill':
            return (self.shear('Voigt') + self.shear('Reuss')) / 2
        
        elif style == 'Voigt':
            c = self.Cij
            return ( (c[0,0] + c[1,1] + c[2,2]) - (c[0,1] + c[1,2] + c[0,2]) + 3*(c[3,3] + c[4,4] + c[5,5]) ) / 15
        
        elif style == 'Reuss':
            s = self.Sij
            return 15 / ( 4*(s[0,0] + s[1,1] + s[2,2]) - 4*(s[0,1] + s[1,2] + s[0,2]) + 3*(s[3,3] + s[4,4] + s[5,5]) )
        
        else:
            raise ValueError('Unknown estimate style')


# coding: utf-8
# cython: language_level=3

# Standard Python libraries
import warnings

# http://www.numpy.org/
import numpy as np

# atomman imports
from .. import NeighborList
from ..tools import axes_check, aslist

cimport cython

from libc.math cimport cos, pi, sqrt, fabs

class Strain():

    def __init__(self, system, neighbors=None, cutoff=None, p_vectors=None,
                 theta_max=27, axes=None, basesystem=None, baseneighbors=None):
        """
        Class initializer.  Allows for the current and reference state to be
        specified for performing the calculations.
        
        Parameters
        ----------
        system : atomman.System
            The atomic system to compute the per-atom strain properties and Nye
            tensor for.
        neighbors : atomman.NeighborList, optional
            The neighbor list associated with system to use.  Either neighbors
            or cutoff must be given, or system must have a neighbors attribute.
        cutoff : float
            Cutoff distance for computing a neighbor list for system.  Either
            neighbors or cutoff must be given, or system have a neighbors
            attribute.
        p_vectors : array-like object, optional
            List(s) of radial distance vectors between each atom and its nearest
            neighbors in a perfect crystal setting.  If one list of p_vectors is
            given, then it is applied to all atoms.
        axes : array-like object, optional
            3x3 array of right-handed orthogonal axes to transform the given p
            vectors.  Only needed if the orientation of the p vectors differs
            from the system's orientation.
        basesystem : atomman.System, optional
            A reference atomic system to use for constructing the p vectors. 
        baseneighbors : atomman.NeighborList, optional
            The neighbor list associated with basesystem to use. If basesystem
            is given, then either baseneighbors or cutoff must be given, or
            basesystem must have a neighbors attribute.
        theta_max : float, optional
            The maximum theta angle in degrees to use when searching for matches
            between p vectors and q vectors.  Optimum values are dependent on the
            crystal structure. Default value is 27, which is the original value
            used for fcc crystals.
        """
        
        self.__system = system
        
        # Neighbor list setup
        if neighbors is not None:
            assert cutoff is None, 'neighbors and cutoff cannot both be given'
            self.__neighbors = neighbors
        elif cutoff is not None:
            self.__neighbors = NeighborList(system=system, cutoff=cutoff)
        elif hasattr(system, 'neighbors'):
            self.__neighbors = system.neighbors
        else:
            raise ValueError('neighbors or cutoff is required')
        
        # p vector setup
        if basesystem is not None:
            assert p_vectors is None, 'basesystem and p_vectors cannot both be given'
            self.build_p_vectors(basesystem, neighbors=baseneighbors, cutoff=cutoff)
        elif p_vectors is not None:
            self.set_p_vectors(p_vectors, axes=axes)
        else:
            self.__p_vectors = None
    
        self.theta_max = theta_max
        self.clear_properties()
    
    @property
    def system(self):
        """atomman.System: The system the properties are being computed for."""
        return self.__system
    
    @property
    def p_vectors(self):
        """numpy.NDArray: The per-atom sets of ideal atom positions."""
        return self.__p_vectors
    
    @property
    def theta_max(self):
        """float: The maximum angle in degrees to include in the p-q vector pairings."""
        return self.__theta_max
    
    @theta_max.setter
    def theta_max(self, value):
        if value <= 180 and value > 0:
            self.__theta_max = value
    
    @property
    def neighbors(self):
        """atomman.NeighborList: The list of neighbors for system."""
        return self.__neighbors
    
    @property
    def G(self):
        """numpy.NDArray : The computed per-atom lattice correspondence tensor"""
        if self.__G is None:
            self.solve_G()
        return self.__G
    
    @property
    def strain(self):
        """numpy.NDArray : The computed per-atom strain tensor"""
        if self.__strain is None:
            self.__strain = strain_c(self.G)
        return self.__strain

    @property
    def invariant1(self):
        """numpy.NDArray : The computed per-atom first strain invariant"""
        if self.__invariant1 is None:
            self.__invariant1 = invariant1_c(self.strain)
        return self.__invariant1
        
    @property
    def invariant2(self):
        """numpy.NDArray : The computed per-atom second strain invariant"""
        if self.__invariant2 is None:
            self.__invariant2 = invariant2_c(self.strain)
        return self.__invariant2
        
    @property
    def invariant3(self):
        """numpy.NDArray : The computed per-atom third strain invariant"""
        if self.__invariant3 is None:
            self.__invariant3 = invariant3_c(self.strain)
        return self.__invariant3

    @property
    def rotation(self):
        """numpy.NDArray : The computed per-atom rotation tensor"""
        if self.__rotation is None:
            self.__rotation = rotation_c(self.G)
        return self.__rotation
        
    @property
    def angularvelocity(self):
        """numpy.NDArray : The computed per-atom angular velocity"""
        if self.__angularvelocity is None:
            self.__angularvelocity = angularvelocity_c(self.rotation)
        return self.__angularvelocity

    @property
    def nye(self):
        """numpy.NDArray : The computed per-atom Nye tensor"""
        if self.__nye is None:
            self.solve_nye()
        return self.__nye

    def clear_properties(self):
        """
        Clears all computed properties. Allows for the values to be recomputed
        using different settings.
        """
        self.__G = None
        self.__strain = None
        self.__invariant1 = None
        self.__invariant2 = None
        self.__invariant3 = None
        self.__angularvelocity = None
        self.__rotation = None
        self.__nye = None

    def save_to_system(self, properties=None):
        """
        Saves the computed per-atom strain properties to the given system.

        Parameters
        ----------
        properties : str or list, optional
            One or more properties.  If not given, will include strain,
            invariant1, invariant2, invariant3, angularvelocity and nye.
        """
        defaultkeys = ['strain', 'invariant1', 'invariant2', 'invariant3',
                       'angularvelocity', 'nye']
        allkeys = ['G', 'rotation'] + defaultkeys

        if properties is None:
            properties = defaultkeys
        else:
            properties = aslist(properties)
        for p in properties:
            assert p in allkeys, 'unknown property ' + p
            self.system.atoms.view[p] = getattr(self, p)

    def asdict(self, properties=None):
        """
        Returns a dictionary containing the computed per-atom strain properties.
        This corresponds to the results from the old nye_tensor function.

        Parameters
        ----------
        properties : str or list, optional
            One or more properties.  If not given, will include strain,
            invariant1, invariant2, invariant3, angularvelocity and nye.

        Returns
        -------
        dict
            Containing each of the computed properties.
        """
        defaultkeys = ['strain', 'invariant1', 'invariant2', 'invariant3',
                       'angularvelocity', 'nye']
        allkeys = ['G', 'rotation'] + defaultkeys
        results = {}
        if properties is None:
            properties = defaultkeys
        else:
            properties = aslist(properties)
        
        results = {}
        for p in properties:
            assert p in allkeys, 'unknown property ' + p
            results[p] = getattr(self, p)
        
        return results

    def set_p_vectors(self, p_vectors, axes=None):
        """
        Sets the p vectors for each atom according to a single p vector set to use
        for all atoms or a list of p vector sets defined for each atom individually.

        Parameters
        ----------
        p_vectors : array-like object
            List(s) of radial distance vectors between each atom and its nearest
            neighbors in a perfect crystal setting.  If one list of p_vectors is
            given, then it is applied to all atoms.
        axes : array-like object, optional
            3x3 array of right-handed orthogonal axes.  If given, will be used to
            transform the p_vectors before computing the Nye tensor.
        """
        system = self.system
        
        # Broadcast a single p_vectors list to all atoms
        if len(p_vectors) == 1:
            p_vectors = np.broadcast_to(p_vectors, (system.natoms, len(p_vectors[0]), 3))
        elif len(p_vectors) != system.natoms:
            p_vectors = np.broadcast_to(p_vectors, (system.natoms, len(p_vectors), 3))
        else:
            for i in range(len(p_vectors)):
                p_vectors[i] = np.asarray(p_vectors[i])
                if p_vectors[i].ndim == 1:
                    p_vectors[i] = np.array([p_vectors[i]])
            p_vectors = np.asarray(p_vectors)

        # Transform p_vectors if axes is given
        if axes is not None:
            p_vectors = np.inner(p_vectors, axes_check(axes))
            
        self.__p_vectors = p_vectors
        
    def build_p_vectors(self, basesystem, neighbors=None, cutoff=None):
        """
        Builds the p vectors for each atom based on a reference system.

        Parameters
        ----------
        basesystem : atomman.system
            The base/reference system to use.  This should be a defect-free
            perfect crystal system with atom ids directly corresponding to atoms
            in any system that you want to analyze with the Nye tensor.
        neighbors : atomman.NeighborList, optional
            The neighbor list associated with system to use.  Either neighbors
            or cutoff must be given, or system must have a neighbors attribute.
        cutoff : float
            Cutoff distance for computing a neighbor list for system.  Either
            neighbors or cutoff must be given, or system have a neighbors
            attribute.
        """

        # Neighbor list setup
        if neighbors is not None:
            assert cutoff is None, 'neighbors and cutoff cannot both be given'
        elif cutoff is not None:
            neighbors = NeighborList(system=basesystem, cutoff=cutoff)
        elif hasattr(basesystem, 'neighbors'):
            neighbors = basesystem.neighbors
        else:
            raise ValueError('neighbors or cutoff is required')

        p = []
        for i in range(basesystem.natoms):
            p.append(basesystem.dvect(i, neighbors[i]))

        self.__p_vectors = np.asarray(p)

    def solve_G(self, theta_max=None):
        """
        Computes the lattice correspondence tensor, G, by comparing all neighbor
        vectors q in the current system to ideal reference neighbor vectors p.
        
        Parameters
        ----------
        theta_max : float, optional
            Allows for the theta_max value set during the class initialization
            to be changed.  The maximum theta angle in degrees to use when
            searching for matches between p vectors and q vectors.  Optimum
            values are dependent on the crystal structure. 
        """
        # p vector setup
        p_vectors = self.p_vectors
        if p_vectors is None:
            raise ValueError('Cannot solve until p_vectors are set')

        # Theta_max setup
        if theta_max is not None:
            self.theta_max = theta_max
        cos_theta_max = cos(self.theta_max * pi / 180.0)

        # system and neighbors setup
        system = self.system
        neighbors = self.neighbors
        maxcoord = neighbors.coord.max()
        natoms = system.natoms

        # Initialize variables
        self.clear_properties()
        G = np.empty((natoms, 3, 3))
        P = np.empty((maxcoord, 3))
        Q = np.empty((maxcoord, 3))

        # Loop over all atoms
        for i in range(natoms):
            # Get actual p and q for atom i
            p = p_vectors[i]
            q = system.dvect(i, neighbors[i])
            
            # Get the matched P and Q
            n = match_pq(p, q, cos_theta_max, P, Q)
            
            # Compute lattice correspondence tensor, G, from P and Q
            if n == 0:
                G[i] = np.identity(3)
                warnings.warn('An atom lacks pair sets. Check neighbor list')
            else:
                G[i] = np.linalg.lstsq(Q[:n], P[:n], rcond=None)[0]

        self.__G = G

    def solve_nye(self):
        """
        Computes the Nye tensor based on the lattice correspondence tensor G
        and the current atomic positions.
        """
        G = self.G
        cdef double[:,:,::1] G_view = G
        system = self.system
        nlist = self.neighbors.nlist
        cdef long long[:,::1] nlist_view = nlist
        
        cdef Py_ssize_t i, x, y, z
        
        # Variable setup and shortcuts
        natoms = G.shape[0]
        cdef long long maxcoord = 0
        for i in range(natoms):
            if maxcoord < nlist_view[i, 0]:
                maxcoord = nlist_view[i, 0]
        
        nye = np.empty((natoms, 3, 3))
        cdef double [:,:,::1] nye_view = nye
        gradG = np.empty((3, 3, 3))
        cdef double [:,:,::1] gradG_view = gradG
        
        gG = np.empty((3,3))
        cdef double [:,::1] gG_view = gG
        
        dG = np.empty((maxcoord, 3, 3))
        cdef double [:,:,::1] dG_view = dG
        
        Q = np.empty((maxcoord, 3))
        cdef double [:,::1] Q_view = Q
        
        # Construct the gradient tensor of G, gradG for each atom
        for i in range(natoms):
            
            c = nlist_view[i, 0]
            Q[:c] = system.dvect(i, nlist[i, 1:c+1])
            
            dG_c(G_view, nlist_view, dG_view, i)
            
            for x in range(3):
                gG[:] = np.linalg.lstsq(Q[:c], dG[:c,x,:], rcond=None)[0]
                
                for y in range(3):
                    for z in range(3):
                        gradG_view[x,y,z] = gG_view[z,y]

            nye_c(gradG_view, nye_view, i)
            
        self.__nye = nye

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef match_pq(double[:,::1] p, 
              double[:,::1] q,
              double cos_theta_max,
              double[:,::1] P, 
              double[:,::1] Q):
    """
    Takes p and q neighbor vectors and finds corresponding pairs P and Q based
    on the angles beween vectors (and interatomic spacing).
    """
    # Parameter initialization
    pnum = p.shape[0]
    qnum = q.shape[0]
    cdef long long[:] qp_pairs = np.empty(qnum, dtype='int64')
    cdef double [:] pmag = np.empty(pnum)
    cdef double [:] qmag = np.empty(qnum)
    
    cdef double r1 = 1e16
    cdef double cos_theta_min, jrad, krad
    
    cdef Py_ssize_t j, k, n, x, y, z
    
    # Compute magnitudes of q and set all qp_pairs to -1
    for j in range(qnum):
        qmag[j] = sqrt(q[j,0] * q[j,0] + q[j,1] * q[j,1] + q[j,2] * q[j,2])
        qp_pairs[j] = -1
        
    # Compute magnitudes of p and find the shortest r1
    for k in range(pnum):
        pmag[k] = sqrt(p[k,0] * p[k,0] + p[k,1] * p[k,1] + p[k,2] * p[k,2])
        if pmag[k] < r1:
            r1 = pmag[k]
    
    # Loop over all q
    for j in range(qnum):
        cos_theta_min = cos_theta_max
    
        # Loop over all p
        for k in range(pnum):
            
            # Compute cos theta between p and q
            cos_theta = (q[j,0] * p[k,0] + q[j,1] * p[k,1] + q[j,2] * p[k,2]) / (qmag[j] * pmag[k])
            
            # Select q-p pair with the smallest theta angle (largest cos theta)
            if cos_theta > cos_theta_min:
                cos_theta_min = cos_theta
                qp_pairs[j] = k
                
        # Check if the best p is already assigned to another q
        if qp_pairs[j] >= 0:
            for k in range(j):
                if qp_pairs[j] == qp_pairs[k]:
                    jrad = fabs(r1 - qmag[j])
                    krad = fabs(r1 - qmag[k])

                    # Remove the p-q pair that is farther from r1
                    if jrad < krad:
                        qp_pairs[k]=-1
                    else:
                        qp_pairs[j]=-1 
        
    # Construct reduced P, Q matrices from p-q pairs
    n = 0
    for j in range(qnum):
        if qp_pairs[j] >= 0:
            for x in range(3):
                Q[n,x] = q[j,x]
                P[n,x] = p[qp_pairs[j],x]
            n+=1
            
    # Return number of matching P-Q pairs 
    return n

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef strain_c(const double[:,:,::1] G):
    """
    Computes the per-atom strain tensor from the lattice correspondence tensor G
    """
    # Parameter initialization 
    natoms = G.shape[0]
    ident = np.identity(3)
    cdef double [:,::1] identity = ident
    
    strain = np.empty((natoms, 3, 3))
    cdef double [:,:,::1] strain_view = strain
    cdef Py_ssize_t i, j, k
    
    # ε_jk = ((I_jk - G_jk) + (I_kj - G_kj)) / 2
    for i in range(natoms):
        for j in range(3):
            for k in range(3):
                strain_view[i,j,k] = ((identity[j,k] - G[i,j,k]) + (identity[k,j] - G[i,k,j])) / 2.
    return strain

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef invariant1_c(const double [:,:,::1] strain):
    """
    Computes the per-atom first strain invariant from the strain tensor
    """
    # Parameter initialization 
    natoms = strain.shape[0]
    
    invar = np.empty(natoms)
    cdef double [::1] invar_view = invar
    cdef Py_ssize_t i
    
    # I1 = trace(ε_ij)
    for i in range(natoms):
        invar_view[i] = strain[i,0,0] + strain[i,1,1] + strain[i,2,2]
    
    return invar

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef invariant2_c(const double [:,:,::1] strain):
    """
    Computes the per-atom second strain invariant from the strain tensor
    """
    # Parameter initialization 
    natoms = strain.shape[0]
    
    invar = np.empty(natoms)
    cdef double [::1] invar_view = invar
    cdef Py_ssize_t i
    
    # I2 = trace(ε_ij)
    for i in range(natoms):
        invar_view[i] = ( strain[i,0,0] * strain[i,1,1] 
                        + strain[i,0,0] * strain[i,2,2] 
                        + strain[i,1,1] * strain[i,2,2] 
                        - strain[i,0,1] * strain[i,1,0] 
                        - strain[i,0,2] * strain[i,2,0]
                        - strain[i,1,2] * strain[i,2,1])
    
    return invar

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef invariant3_c(const double [:,:,::1] strain):
    """
    Computes the per-atom third strain invariant from the strain tensor
    """
    # Parameter initialization 
    natoms = strain.shape[0]
    
    invar = np.empty(natoms)
    cdef double [::1] invar_view = invar
    cdef Py_ssize_t i
    
    # I3 = det(ε_ij)
    for i in range(natoms):
        invar_view[i] = ( strain[i,0,0] * ( strain[i,1,1] * strain[i,2,2]
                                          - strain[i,1,2] * strain[i,2,1]) 
                        - strain[i,0,1] * ( strain[i,1,0] * strain[i,2,2]
                                          - strain[i,1,2] * strain[i,2,0]) 
                        + strain[i,0,2] * ( strain[i,1,0] * strain[i,2,1]
                                          - strain[i,1,1] * strain[i,2,0]))
    
    return invar

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef rotation_c(const double[:,:,::1] G):
    """
    Computes the per-atom rotation tensor from the lattice correspondence tensor G
    """
    # Parameter initialization 
    natoms = G.shape[0]
    ident = np.identity(3)
    cdef double [:,::1] identity = ident
    
    rotation = np.empty((natoms, 3, 3))
    cdef double [:,:,::1] rot_view = rotation
    cdef Py_ssize_t i, j, k
    
    # rot_jk = ((I_jk - G_jk) - (I_kj - G_kj)) / 2
    for i in range(natoms):
        for j in range(3):
            for k in range(3):
                rot_view[i,j,k] = ((identity[j,k] - G[i,j,k]) - (identity[k,j] - G[i,k,j])) / 2.
    return rotation

@cython.boundscheck(False)
@cython.wraparound(False) 
cdef angularvelocity_c(const double [:,:,::1] rotation):
    """
    Computes the per-atom angular velocity from the rotation tensor
    """
    # Parameter initialization 
    natoms = rotation.shape[0]
    
    angvel = np.empty(natoms)
    cdef double [::1] angvel_view = angvel
    cdef Py_ssize_t i
    
    # a = sqrt(rot_12^2 + rot_13^2 + rot_23^2)
    for i in range(natoms):
        angvel_view[i] = sqrt( rotation[i,0,1] * rotation[i,0,1]
                             + rotation[i,0,2] * rotation[i,0,2]
                             + rotation[i,1,2] * rotation[i,1,2] )
    return angvel   



@cython.boundscheck(False)
@cython.wraparound(False) 
cdef dG_c(const double[:,:,::1] G, const long long[:,::1] nlist, double[:,:,::1] dG, Py_ssize_t i):
    """
    Computes the change in G between an atom and each of its neighbors
    """
    coord = nlist[i, 0]    
    cdef Py_ssize_t j, x, y
    
    # Calculate change in G between all neighbors
    for j in range(coord):
        for x in range(3):
            for y in range(3):
                dG[j, x, y] = G[nlist[i, j+1], x, y] - G[i, x, y]
    
@cython.boundscheck(False)
@cython.wraparound(False)
cdef nye_c(double [:,:,::1] gradG, double [:,:,::1] nye, Py_ssize_t i):
    """
    Computes the per-atom Nye tensor from the calculated grad(G).
    """
    nye[i,0,0] = gradG[1,0,2] - gradG[2,0,1]
    nye[i,0,1] = gradG[1,1,2] - gradG[2,1,1]
    nye[i,0,2] = gradG[1,2,2] - gradG[2,2,1]
    nye[i,1,0] = gradG[2,0,0] - gradG[0,0,2]
    nye[i,1,1] = gradG[2,1,0] - gradG[0,1,2]
    nye[i,1,2] = gradG[2,2,0] - gradG[0,2,2]
    nye[i,2,0] = gradG[0,0,1] - gradG[1,0,0]
    nye[i,2,1] = gradG[0,1,1] - gradG[1,1,0]
    nye[i,2,2] = gradG[0,2,1] - gradG[1,2,0]


# coding: utf-8

# http://www.numpy.org/
import numpy as np

def axes_check(axes, tol=1e-8):
    """
    Checks that given axes are orthogonal and right-handed.
    
    Parameters
    ----------
    axes : list of list or np.ndarray
        A 3x3 array of axes vectors.
    tol : float, optional
        Tolerance to use in checking if axes are orthogonal and right-handed.
        Default value is 1e-8.
    
    Returns
    -------
    np.ndarray
        A 3x3 array of the corresponding unit vectors of axes.
    
    Raises
    ------
    ValueError
        If axes are not orthogonal or right-handed.
    """
    # Convert to numpy.ndarray if needed
    axes = np.asarray(axes)
    assert axes.shape == (3,3), 'Invalid axes shape'
    
    # Compute unit axes
    uaxes = (axes.T / np.linalg.norm(axes, axis=1)).T
    
    # Check if orthogonal
    if not np.allclose(np.dot(uaxes, uaxes.T), np.identity(3), atol=tol):
        raise ValueError('axes are not orthogonal')
    
    # Check if right-handed
    if not np.allclose(np.cross(uaxes[0], uaxes[1]), uaxes[2], atol=tol):
        raise ValueError('axes are not right-handed')
    
    return uaxes


# coding: utf-8
# Standard Python libraries
from collections import OrderedDict

# atomman imports
import atomman.unitconvert as uc
from .process_prop_info import process_prop_info
from ...tools import indexstr

def dump(system, f=None, prop_name=None, table_name=None, shape=None,
         unit=None, dtype=None, prop_info=None, header=False,
         float_format ='%.13f', return_prop_info=False, extra=None):
    """
    Converts a system's atoms' values to a string table.
    
    Parameters
    ----------
    system : atomman.System
        An atomman representation of a system.
    f : str or file-like object, optional
        File path or file-like object to write the content to.  If not given,
        then the content is returned as a str.
    prop_name : list, optional
        The Atoms properties to include.  Must be given if prop_info is not.
    table_name : list, optional
        The table column name(s) that correspond to each prop_name.  If not
        given, the table_name values will be based on the prop_name values.
    shape : list, optional
        The shape of each per-atom property.  If not given, will be inferred
        from the length of each table_name value.
    unit : list, optional
        Lists the units for each prop_name as stored in the table.  For a
        value of None, no conversion will be performed for that property.  For
        a value of 'scaled', the corresponding table values will be taken in
        box-scaled units.  If not given, all unit values will be set to None
        (i.e. no conversions).
    dtype : list, optional
        Allows for the data type of each property to be explicitly given.
        Values of None will infer the data type from the corresponding
        property values.  If not given, all values will be None.
    prop_info : list of dict, optional
        Structured form of property conversion parameters, in which each
        dictionary in the list corresponds to a single atoms property.  Each
        dictionary must have a 'prop_name' field, and can optionally have
        'table_name', 'shape', 'unit', and 'dtype' fields.
    header : bool, optional
        Flag indicating whether to include the column names in the outputted
        table.  Default value is False (no column names).
    float_format : str, optional
        c-style formatting string for floating point values.  Default value is
        '%.13f'.
    return_prop_info : bool, optional
        Flag indicating if the filled-in prop_info is to be returned.  Having
        this allows for 1:1 load/dump conversions.  Default value is False
        (prop_info is not returned).
    extra : dict, optional
        Allows extra per-atom data that is not part of the System to be
        included in the generated table.  Useful when the per-atom data only
        has meaning in the tabular format and should not be added to System.
        
    Returns
    -------
    str
        The generated data table.  Only returned if fp is None.
    """
    # Set parameters
    natoms = system.natoms
    key_rename = OrderedDict()
    
    # Set default values
    if prop_info is None:
        if prop_name is None:
            prop_name = system.atoms_prop()
        
        if shape is None and table_name is None:
            shape = []
            for name in prop_name:
                shape.append(system.atoms.view[name].shape[1:])
    
    # Process conversion parameters
    prop_info = process_prop_info(prop_name=prop_name, table_name=table_name,
                                  shape=shape, unit=unit, dtype=dtype,
                                  prop_info=prop_info)
    
    
    # Build list of properties to scale
    scale = []
    for prop in prop_info:
        if prop['unit'] == 'scaled':
            scale.append(prop['prop_name'])
            prop['unit'] = None
    
    # Transform to dataframe
    df = system.atoms_df(scale)
    
    # Add a_id values
    df['a_id'] = range(1, natoms+1)
    
    # Loop over all properties
    for prop in prop_info:
        pname = prop['prop_name']
        
        # loop over all table names and property indexes
        for tname, (index, istr) in zip(prop['table_name'], # pylint: disable=unused-variable
                                        indexstr(prop['shape'])):
            
            # Build name change dict
            key_rename[pname + istr] = tname
            
            # Convert units if needed
            if prop['unit'] is not None:
                df[pname + istr] = uc.get_in_units(df[pname + istr], prop['unit'])
    
    # Rename and reorganize
    df = df.rename(columns=key_rename)[list(key_rename.values())]
  
    # Add extra content if given
    if extra is not None:
        for key, value in extra.items():
            df[key] = value

    # Generate table
    sep = ' '
    table = df.to_csv(path_or_buf=f,
                      sep=sep,
                      index=None,
                      header=header,
                      float_format=float_format,
                      encoding='ascii',
                      line_terminator='\n',
                      )
    
    returns = []
    
    if table is not None:
        returns.append(table)
    
    if return_prop_info is True:
        returns.append(prop_info)
        
    if len(returns) == 1:
        return returns[0]
    elif len(returns) > 1:
        return tuple(returns)
    
    return


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# https://www.scipy.org/
from scipy.interpolate import griddata

# https://matplotlib.org/
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
from matplotlib import cm

# atomman imports
import atomman.unitconvert as uc
from ..tools import axes_check, vect_angle

__all__ = ['interpolate_contour']

def interpolate_contour(system, name, property=None, index=None, magnitude=False,
                        plotxaxis='x', plotyaxis='y', xlim=None, ylim=None,
                        zlim=None, xbins=200, ybins=200, dots=True, czero=True,
                        save=False, show=True, length_unit='angstrom',
                        property_unit=None, cmap='jet'):
    """
    Creates a contour plot of a system's per-atom properties by interpolating
    properties between atoms.
    
    Parameters
    ----------
    system : atomman.System
        The system with the per-atom property that you want to plot.
    name : str
        The name of the per-atom property that you want to plot.
    property : array-like object, optional
        Values for the per-atom property to plot.  If not given, values will
        be taken as the "name" property of system.
    index : int or tuple, optional
        Specifies which component of a multidimensional property to plot.  Not
        needed if the property is scalar.
    magnitude : bool, optional
        If True, plots the per-atom magnitude of a vector property.  Cannot be
        combined with index.  Default value is False.
    plotxaxis : str or array-like object, optional
        Indicates the Cartesian direction associated with the system's atomic
        coordinates to align with the plotting x-axis.  Values are either 3D
        unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
        directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
        is 'x' = [1, 0, 0].
    plotyaxis : str or array-like object, optional
        Indicates the Cartesian direction associated with the system's atomic
        coordinates to align with the plotting y-axis.  Values are either 3D
        unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
        directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
        is 'y' = [0, 1, 0].
    xlim : tuple, optional
        The minimum and maximum coordinates along the plotting x-axis to
        include in the fit.  Values are taken in the specified length_unit.
        If not given, then the limits are set based on min and max atomic
        coordinates along the plotting axis.
    ylim : tuple, optional
        The minimum and maximum coordinates along the plotting y-axis to
        include in the fit.  Values are taken in the specified length_unit.
        If not given, then the limits are set based on min and max atomic
        coordinates along the plotting axis.
    zlim : tuple, optional
        The minimum and maximum coordinates normal to the plotting axes
        (i.e. plotxaxis X plotyaxis) to include in the fit.  Values are taken
        in the specified length_unit.  If not given, then the limits are set
        based on min and max atomic coordinates along the axis.    
    xbins : int, optional
        Specifies the number of interpolation bins to use along the plotting 
        x-axis.  Default value is 200.
    ybins : int, optional
        Specifies the number of interpolation bins to use along the plotting
        y-axis.  Default value is 200.
    dots : bool, optional
        If True, then the positions of the atoms are shown as circles.
        Default value is True.
    czero : bool, optional
        If True, the range of property values will be centered around zero,
        i.e. cmax = -cmin.  If False, cmax and cmin will be independently
        selected using the property values.  Default value is True.
    save : bool, optional
        If True, the generated plot will be saved to "name.png".  Default
        value is False.
    show : bool, optional
        If True, matplotlib.pyplot.show() is called.  Default value is True.
    length_unit : str, optional
        The units of length to use for the plotting x- and y-axes.  Default
        value is 'angstrom'.
    property_unit : str or None, optional
        The units to use for the property value being plotted.  Default value
        is None, in which no unit conversion is applied.
    cmap : str, optional
        The name of the matplotlib colormap to use.  Default value is 'jet'.
    
    Returns
    -------
    intsum : float
        The area integrated sum of the property over the plotted region.
    avsum : float
        The average property value taken across all plotting bins.
    """
    
    # Give numeric values for str plot axis terms
    if plotxaxis == 'x':
        plotxaxis = [1.0,0.0,0.0]
    elif plotxaxis == 'y':
        plotxaxis = [0.0,1.0,0.0]
    elif plotxaxis == 'z':
        plotxaxis = [0.0,0.0,1.0]
    
    if plotyaxis == 'x':
        plotyaxis = [1.0,0.0,0.0]
    elif plotyaxis == 'y':
        plotyaxis = [0.0,1.0,0.0]
    elif plotyaxis == 'z':
        plotyaxis = [0.0,0.0,1.0]
    
    # Build transformation matrix, T, from plot axes.
    plotxaxis = np.asarray(plotxaxis, dtype='float64')
    plotyaxis = np.asarray(plotyaxis, dtype='float64')
    T = axes_check([plotxaxis, plotyaxis, np.cross(plotxaxis, plotyaxis)])
    
    # Extract positions and transform using T
    pos = uc.get_in_units(np.inner(system.atoms.pos, T), length_unit)
    
    # Set plot limits
    if xlim is None:
        xlim = (pos[:, 0].min(), pos[:, 0].max())
    if ylim is None:
        ylim = (pos[:, 1].min(), pos[:, 1].max())
    if zlim is None:
        zlim = (pos[:, 2].min(), pos[:, 2].max())
    
    # Extract property values
    if property is None:
        property = system.atoms.view[name]
    
    # Handle index
    if index is not None:
        assert magnitude is False, 'index and magnitude cannot be combined'
        if isinstance(index, (int, np.integer)):
            index = [index]
        else:
            index = list(index)
        for i in index:
            name += '[' + str(i+1) + ']'
        property = property[tuple([Ellipsis] + index)]
    
    # Handle magnitude
    elif magnitude is True:
        property = np.linalg.norm(property, axis=1)
        name += '_mag'
    
    assert property.shape == (system.natoms, ), 'property to plot must be a per-atom scalar'
    
    # Identify only atoms in xlim, ylim, zlim
    in_bounds = ((pos[:,0] > xlim[0] - 5.) &
                 (pos[:,0] < xlim[1] + 5.) &
                 (pos[:,1] > ylim[0] - 5.) &
                 (pos[:,1] < ylim[1] + 5.) &
                 (pos[:,2] > zlim[0]) &
                 (pos[:,2] < zlim[1]))
    
    # Extract plotting coordinates and values
    x = pos[in_bounds, 0]
    y = pos[in_bounds, 1]
    v = uc.get_in_units(property[in_bounds], property_unit)
    
    # Generate interpolation grid
    grid, xedges, yedges = grid_interpolate_2d(x, y, v, xbins=xbins, ybins=ybins, range=[xlim, ylim])
    
    # Compute intsum and avsum values
    intsum = np.sum(grid)
    avsum = intsum / (xbins-1) / (ybins-1)
    intsum = intsum * (xlim[1] - xlim[0]) / (xbins-1) * (ylim[1] - ylim[0]) / (ybins-1)
    
    # Generate a pretty figure of the grid
    fig = prettygrid(grid, xedges, yedges, cmap=cmap, propname=name, czero=czero)
    
    # Add dots
    if dots:
        adddots(x, y, xedges, yedges)
    
    # Save, show and close figure
    if save is True:
        plt.savefig(name + '.png', dpi=800)
    if show is True:
        plt.show()
    plt.close(fig)
    
    return intsum, avsum
    
def grid_interpolate_2d(x, y, v, xbins=50, ybins=50, range=None):
    """
    Generates 2D grid of property values by interpolating between measured
    values.
    
    Parameters
    ----------
    x : array-like object
        The list of x-coordinate values.
    y : array-like object
        The list of y-coordinate values.
    v : array-like object
        The list of property values associated with each (x, y) set.
    xbins : int, optional
        The number of bins to use in interpolating between the x coordinates.
        Default value is 50.
    ybins : int, optional
        The number of bins to use in interpolating between the y coordinates.
        Default value is 50.
    range : list, tuple or array-like object
        2x2 list of the [[xmin, xmax], [ymin, ymax]] coordinates for the bins.
        If not given, will use [[x.min(), x.max()], [y.min(), y.max()]].
    
    Returns
    -------
    grid : numpy.ndarray
        The interpolated values in a grid map array.
    xedges : list
        The min, max values of x associated with the grid.
    yedges : list
        The min, max values of y associated with the grid.
    """
    # Handle range and bins options
    if range is None:
        range=[[x.min() ,x.max()], [y.min(), y.max()]]
    
    # Generate 1D interpolation points
    xi = np.linspace(range[0][0], range[0][1], num=xbins)
    yi = np.linspace(range[1][0], range[1][1], num=ybins)
    
    # Generate 2D grid points
    x0, y0 = np.meshgrid(xi, yi)
    
    # Interpolate values to grid points
    grid = griddata((x, y), v, (x0, y0))
    
    return grid, range[0], range[1]

def prettygrid(grid, xedges, yedges, cmap='jet', propname='', czero=True,
               scale=1):
    """
    Generates pretty-looking 2D image maps for grid data using matplotlib.
    
    Parameters
    ----------
    grid : array-like object
        The grid of property values to plot
    xedges : list
        The min, max values of x associated with the grid.
    yedges : list
        The min, max values of y associated with the grid.
    cmap : str, optional
        The name of the matplotlib colormap to use.  Default value is 'jet'.
    propname : str, optional
        Name of the property being printed.  This is used as the plot title.
    czero : bool, optional
        If True, the range of property values will be centered around zero,
        i.e. cmax = -cmin.  If False, cmax and cmin will be independently
        selected using the property values.  Default value is True.
    scale : int, optional
        Scaling factor to apply to min/max property values.  Default value is
        1 (no scaling).
    
    Returns
    -------
    matplotlib.Figure
        The generated figure object.
    """
    
    # Get cmap from str name
    if isinstance(cmap, str):
        cmap = cm.get_cmap(cmap)
    
    # Set vmin = -vmax if czero is True
    if czero is True:
        vmax = abs(grid).max()
        if vmax != 0.0:
            vrounder = np.floor(np.log10(vmax))
            vmax = np.around(2 * vmax / 10.**vrounder) * 10.**vrounder / 2.
        else:
            vmax = 1e-15
        vmin = -vmax
    
    # Set vmin, vmax if czero is False
    else:
        vmax = grid.max()
        vmin = grid.min()

        if abs(grid).max() != 0.0:
            vrounder = np.floor(np.log10(grid.max() - grid.min()))

            vmax = np.around(2 * vmax / 10.**vrounder) * 10.**vrounder / 2.
            vmin = np.around(2 * vmin / 10.**vrounder) * 10.**vrounder / 2.

            if vmax == vmin:
                if vmax > 0:
                    vmin = 0
                else:
                    vmax = 0
        
        else:
            vmax = 1e-15
            vmin = -1e-15
    
    # Scale values if needed
    vmin *= scale
    vmax *= scale
    
    # Generate plot
    fig = plt.figure(figsize=(7.7, 7), dpi=72)
    plt.imshow(grid, extent = [xedges[0], xedges[-1], yedges[-1], yedges[0]], 
               cmap=cmap, norm=plt.Normalize(vmax=vmax, vmin=vmin))
    
    # Add pretty colorbar
    vticks = np.linspace(vmin, vmax, 11, endpoint=True)
    cbar = plt.colorbar(fraction=0.05, pad = 0.05, ticks=vticks)
    cbar.ax.tick_params(labelsize=15)
   
    # Make axis values and title pretty
    plt.xlim(xedges[0], xedges[-1])
    plt.xticks(size=15)
    plt.ylim(yedges[0], yedges[-1])
    plt.yticks(size=15)
    plt.title(propname, size=30)
    
    return fig
    
def adddots(x, y, xedges, yedges):
    """
    Overlays circles onto an active 2D plot to show actual atomic positions.
    
    Parameters
    ----------
    x : array-like object
        List of x-coordinates for the atoms/dots.
    y : array-like object
        List of y-coordinates for the atoms/dots.
    xedges : list
        The [xmin, xmax] values associated with the plot.
    yedges : list
        The [ymin, ymax] values associated with the plot.
    """
    # Get current plotting axis
    ax = plt.gca()
    
    # Set linewidth based on system dimensions
    syswidth = max([abs(xedges[-1]-xedges[0]), abs(yedges[-1]-yedges[0])])
    linewidth = 60. / syswidth
    
    # Add circles at each (x,y) position
    for xi, yi in zip(x, y):
        point = mpatches.Circle((xi, yi), 0.3, ls='solid', lw=linewidth)
        ax.add_artist(point)
        point.set_facecolor('none')
        point.set_edgecolor('k')


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from .nlist import nlist # pylint: disable=no-name-in-module
from ..tools import uber_open_rmode

class NeighborList(object):
    """Class that finds and stores the neighbor atoms for a system."""
    
    def __init__(self, **kwargs):
        """
        Class initializer.  Calls NeighborList.load() if
        'model' is given, otherwise calls NeighborList.build().
        
        Parameters
        ----------
        system : atomman.System, optional
            The system to calculate the neighbor list for. Must be given if
            model is not given.
        cutoff : float, optional
            Radial cutoff distance for identifying neighbors.  Must be given if
            model is not given.
        model : str or file-like object, optional
            Gives the file path or content to load.  If given, no other
            parameters are allowed.
        initialsize : int, optional
            The number of neighbor positions to initially assign to each atom.
            Default value is 20.
        deltasize : int, optional
            Specifies the number of extra neighbor positions to allow each atom
            when the number of neighbors exceeds the underlying array size.
            Default value is 10.
        """
        if 'model' in kwargs:
            model = kwargs.pop('model')
            self.load(model, **kwargs)
        else:
            system = kwargs.pop('system')
            cutoff = kwargs.pop('cutoff')
            self.build(system, cutoff, **kwargs)
    
    @property
    def coord(self):
        """The atomic coordination numbers"""
        return self.__coord
    
    @property
    def nlist(self):
        """numpy.ndarray: The underlying numpy array of coord + neighbor ids"""
        return self.__nlist
    
    def __len__(self):
        """len returns the number of atoms"""
        return len(self.__coord)
    
    def __getitem__(self, key):
        """Get returns the list of neighbors for the specified atom."""
        return self.__neighbors[key, :self.coord[key]]
    
    def build(self, system, cutoff, initialsize=20, deltasize=10):
        """
        Builds the neighbor list for a system.
        
        Parameters
        ----------
        system : atomman.System 
            The system to calculate the neighbor list for.
        cutoff : float
            Radial cutoff distance for identifying neighbors.
        initialsize : int, optional
            The number of neighbor positions to initially assign to each atom.
            Default value is 20.
        deltasize : int, optional
            Specifies the number of extra neighbor positions to allow each atom
            when the number of neighbors exceeds the underlying array size.
            Default value is 10.
        """
        # Call nlist
        self.__nlist = nlist(system, cutoff, initialsize=initialsize, 
                             deltasize=deltasize)
        
        # Split coord and neighbors
        self.__coord = self.__nlist[:, 0]
        self.__neighbors = self.__nlist[:, 1:]
        
    def load(self, model):
        """
        Read in a neighbor list from a file.
        
        Parameters
        ----------
        model : str or file-like object
            Gives the file path or content to load.
        """
        # First pass determines number of atoms and max number of neighbors
        nterms = 0
        natoms = 0
        with uber_open_rmode(model) as fin:
            for line in fin:
                line = line.decode('UTF-8')
                terms = line.split()
                n_n = len(terms)
                if terms[0][0] != '#' and n_n > 0:
                    natoms += 1
                    if n_n > nterms:
                        nterms = n_n
            
            self.__nlist = np.empty((natoms, nterms+1), dtype=int)
            self.__coord = self.__nlist[:, 0]
            self.__neighbors = self.__nlist[:, 1:]
            
            # Second pass gets values
            self.__coord[:] = 0
            fin.seek(0)
            for line in fin:
                line = line.decode('UTF-8')
                terms = line.split()
                if len(terms) > 0 and terms[0][0] != '#':
                    i = int(terms[0])
                    self.__coord[i] = len(terms) - 1
                    for j in range(1, len(terms)):
                        self.__neighbors[i, j-1] = terms[j]
    
    def dump(self, fname):
        """
        Saves the neighbor list to a file.
        
        Parameters
        ----------
        fname : str
            The file name to save the content to.
        """
        with open(fname, 'w') as fp:
            fp.write('# Neighbor list:\n')
            fp.write('# The first column gives an atom index.\n')
            fp.write('# The rest of the columns are the indexes of the identified neighbors.\n')
            for i in range(len(self)):
                fp.write('%i' % i)
                for j in self[i]:
                    fp.write(' %i' % j)
                fp.write('\n')


# coding: utf-8
# Standard Python libraries
import warnings

# http://www.numpy.org/
import numpy as np

# atomman imports
from .. import NeighborList
from ..tools import axes_check

def nye_tensor(system, p_vectors, theta_max = 27, axes=None, neighbors=None,
               cutoff=None):
    """
    Computes strain properties and Nye tensor for a defect containing system.
    
    Parameters
    ----------
    system : atomman.System
        The atomic system to compute the per-atom strain properties and Nye
        tensor for.
    p_vectors : array-like object
        List(s) of radial distance vectors between each atom and its nearest
        neighbors in a perfect crystal setting.  If one list of p_vectors is
        given, then it is applied to all atoms.
    theta_max : float, optional
        The maximum theta angle in degrees to use when searching for matches
        between p vectors and q vectors.  Optimum values are dependent on the
        crystal structure. Default value is 27, which is the original value
        used for fcc crystals.
    axes : array-like object, optional
        3x3 array of right-handed orthogonal axes.  If given, will be used to
        transform the p_vectors before computing the Nye tensor.
    neighbors : atomman.NeighborList, optional
        The neighbor list associated with system to use.  Either neighbors
        or cutoff must be given, or system must have a neighbors attribute.
    cutoff : float
        Cutoff distance for computing a neighbor list for system.  Either
        neighbors or cutoff must be given, or system have a neighbors
        attribute.
    
    Returns
    -------
    dict
        Contains the per-atom properties 'strain', 'strain_invariant_1',
        'strain_invariant_2', 'angular_velocity', and 'Nye_tensor'.
    """
    # Neighbor list setup
    if neighbors is not None:
        assert cutoff is None, 'neighbors and cutoff cannot both be given'
    elif cutoff is not None:
        neighbors = NeighborList(system=system, cutoff=cutoff)
    elif hasattr(system, 'neighbors'):
        neighbors = system.neighbors
    else:
        raise ValueError('neighbors or cutoff is required')
    
    # Broadcast a single p_vectors list to all atoms
    if len(p_vectors) == 1:
        p_vectors = np.broadcast_to(p_vectors, (system.natoms, len(p_vectors[0]), 3))
    elif len(p_vectors) != system.natoms:
        p_vectors = np.broadcast_to(p_vectors, (system.natoms, len(p_vectors), 3))
    
    # Transform p_vectors if axes is given
    if axes is not None:
        p_vectors = np.inner(p_vectors, axes_check(axes))
    
    # Get cos of theta_max
    cos_theta_max = np.cos(theta_max * np.pi / 180)
    
    # Define epsilon array
    eps = np.array([[[ 0, 0, 0],[ 0, 0, 1],[ 0,-1, 0]],
                    [[ 0, 0,-1],[ 0, 0, 0],[ 1, 0, 0]],
                    [[ 0, 1, 0],[-1, 0, 0],[ 0, 0, 0]]])
    
    # Identify largest number of nearest neighbors
    nmax = neighbors.coord.max()
    
    # Initialize variables
    strain = np.empty((system.natoms, 3, 3))
    inv1 = np.empty(system.natoms)
    inv2 = np.empty(system.natoms)
    inv3 = np.empty(system.natoms)
    ang_vel = np.empty(system.natoms)
    nye = np.empty((system.natoms, 3, 3))
    P = np.zeros((nmax, 3))
    Q = np.zeros((nmax, 3))
    G = np.empty((system.natoms, 3, 3))
    gradG = np.empty((3, 3, 3))
    
    # Calculate correspondence tensor, G, and strain data for each atom
    for i in range(system.natoms):
        p = np.asarray(p_vectors[i])
        if p.ndim == 1:
            p = np.array([p])
        p_mags = np.linalg.norm(p, axis=1)
        r1 = p_mags.min()
        
        # Calculate radial neighbor vectors, q
        q = system.dvect(i, neighbors[i])
        if q.ndim == 1:
            q = np.array([q])
        q_mags = np.linalg.norm(q, axis=1)
        
        # Calculate cos_thetas between all p's and q's.
        cos_thetas = (np.dot(p, q.T) /q_mags ).T / p_mags
        
        # Identify best index matches
        index_pairing = cos_thetas.argmax(1)
        
        # Exclude values where theta is greater than theta_max
        index_pairing[cos_thetas.max(1) < cos_theta_max] = -1
        
        # Search for duplicate index_pairings
        #u, u_count = np.unique(index_pairing, return_counts=True)
        
        # Check if the particular p has already been assigned to another q
        for n in range(len(q)):
            if index_pairing[n] >=0:
                for k in range(n):
                    if index_pairing[n] == index_pairing[k]:
                        nrad = abs(r1 - q_mags[n])
                        krad = abs(r1 - q_mags[k])
                        # Remove the p-q pair that is farther from r1
                        if nrad < krad:
                            index_pairing[k]=-1
                        else:
                            index_pairing[n]=-1
        
        # Construct reduced P, Q matrices from p-q pairs
        c = 0
        for n in range(len(q)):
            if index_pairing[n] >= 0:
                Q[c] = q[n]
                P[c] = p[index_pairing[n]]
                c+=1
                
        # Compute lattice correspondence tensor, G, from P and Q
        if c == 0:
            G[i] = np.identity(3)
            warnings.warn('An atom lacks pair sets. Check neighbor list')
        else:
            G[i] = np.linalg.lstsq(Q[:c], P[:c], rcond=None)[0]
        
        # Compute strain properties from G
        strain[i] = ((np.identity(3) - G[i]) + (np.identity(3) - G[i]).T) / 2.
        inv1[i] = strain[i,0,0] + strain[i,1,1] + strain[i,2,2]
        inv2[i] = (strain[i,0,0] * strain[i,1,1] 
                 + strain[i,0,0] * strain[i,2,2] 
                 + strain[i,1,1] * strain[i,2,2] 
                - strain[i,0,1]**2 - strain[i,0,2]**2 - strain[i,1,2]**2)
        inv3[i] = np.linalg.det(strain[i])
        rot = ((np.identity(3) - G[i]) - (np.identity(3) - G[i]).T) / 2.
        ang_vel[i] = (rot[0,1]**2 + rot[0,2]**2 + rot[1,2]**2)**0.5
    
    # Construct the gradient tensor of G, gradG for each atom
    for i in range(system.natoms):
        Q = system.dvect(i, neighbors[i])
        if Q.ndim == 1:
            Q = np.array([Q])
        dG = G[neighbors[i]] - G[i]
        for x in range(3):
            gradG[x,:] = np.linalg.lstsq(Q, dG[:,x,:], rcond=None)[0].T
        
        # Use gradG to calculate the nye tensor
        nye[i] = -1*np.einsum('ijm,ikm->jk', eps, gradG)
    
    return {'strain':strain, 'strain_invariant_1':inv1,
            'strain_invariant_2':inv2, 'strain_invariant_3':inv3,
            'angular_velocity':ang_vel, 'Nye_tensor':nye}


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# https://atztogo.github.io/phonopy/
try:
    import phonopy
    from phonopy.structure.atoms import PhonopyAtoms
    has_phonopy = True
except:
    has_phonopy = False

def dump(system, symbols=None, return_prop=False):
    """
    Convert an atomman.System into a phonopy.structure.atoms.PhonopyAtoms.
    
    Parameters
    ----------
    system : atomman.System
        A atomman representation of a system.
    symbols : tuple, optional
        List of the element symbols that correspond to the atom types.  If not
        given, will use system.symbols if set, otherwise no element content
        will be included.
    return_prop : bool, optional
        Indicates if the extra per-atom properties are to be returned in a
        dictionary.  Default value is False.
    Returns
    -------
    phonopyatoms : phonopy.structure.atoms.PhonopyAtoms
        A phonopy representation of a collection of atoms, based on ase.Atoms.
    prop : dict
        Dictionary containing any extra per-atom properties to include.
        Returned if return_prop is True.
    """
    
    assert has_phonopy, 'phonopy not imported'
    
    # Get box/cell information
    cell = system.box.vects
    pbc = system.pbc
    
    # Get symbols information
    if symbols is None:
        symbols = system.symbols
    symbols = np.asarray(symbols)
    if None in symbols:
        raise ValueError('Symbols needed for all atypes')
    
    # Convert short symbols list to full allsymbols list
    atype = system.atoms.atype
    allsymbols = symbols[atype-1].tolist()
    
    # Get atomic information
    positions = system.atoms.pos
    prop = {}
    for p in system.atoms_prop():
        if p != 'atype' and p != 'pos':
            prop[p] = system.atoms_prop(key=p)
    
    # Build Atoms
    phonopyatoms = PhonopyAtoms(symbols=allsymbols, positions=positions, pbc=pbc, cell=cell)
    
    if return_prop is True:
        return phonopyatoms, prop
    else:
        return phonopyatoms


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from .. import Box, System
from ..tools import vect_angle, miller, ishexagonal

def free_surface_basis(hkl, box=None, cutboxvector='c', maxindex=None,
                       return_hexagonal=None, return_planenormal=False,
                       conventional_setting=None):
    """
    Generates the uvw box vector orientations for a free surface atomic
    system.  In determining the uvw sets, two sets will be in the specified
    hkl plane and one will be out of the plane.  Uses free surface in-plane
    vector determination algorithm by W. Sun and G. Cedar, Surface Science,
    617, 53-59 (2013) to identify two in-plane vectors and the plane normal.
    The shortest in-plane vector is identified, as well as an out-of-plane
    vector close to the plane normal.  The second in-plane vector is then
    selected to be a shortest in-plane vector that is not parallel to the
    first.
    
    Parameters
    ----------
    hkl : array-like object
        The free surface plane to generate expressed in either 3 indices
        Miller (hkl) format or 4 indices Miller-Bravais (hkil) format.
    box : atomman.Box, optional
        The box object associated with the unit cell. Used to identify the
        best uvw set for the out-of-plane box vector.  Default value uses a
        cubic box.
    cutboxvector : str, optional
        Specifies which of the three box vectors corresponds to the
        out-of-plane vector.  Default value is c.
    maxindex : int, optional
        Max uvw index value to use in identifying the best uvw set for the
        out-of-plane vector.  If not given, will use the largest absolute
        index between the given hkl and the initial in-plane vector guesses.
    return_hexagonal : bool, optional
        Flag for indicating if the returned vectors are expressed in Miller
        [uvw] format (False) or Miller-Bravais [uvtw] format (True).  The
        Miller-Bravais format is only allowed if box is in the standard
        hexagonal setting: a=b!=c, alpha=beta=90, gamma=120.  Default value is
        False if hkl is given in the 3 indices Miller (hkl) format and True if
        it is given in the 4 indices Miller-Bravais (hkil) format.
    return_planenormal : bool, optional
        If True, the computed Cartesian plane normal will also be returned.
        Default value is False.
    conventional_setting : str, optional
        Allows for rotations of a primitive unit cell box to be determined from
        (hkl) indices specified relative to a conventional unit cell.  Allowed
        settings: 'p' for primitive (no conversion), 'f' for face-centered,
        'i' for body-centered, and 'a', 'b', or 'c' for side-centered.  Default
        behavior is to perform no conversion, i.e. take (hkl) relative to the
        given box.
    
    Returns
    -------
    uvws : numpy.ndarray
        3x3 array of Miller [uvw] vectors or 3x4 array of Miller-Bravais [uvtw] vectors to rotate the unit cell for a free surface configuration.
    planenormal : numpy.ndarray
        The Cartesian plane normal vector.  Only returned if return_planenormal is True.
        
    Raises
    ------
    ValueError
        If invalid hkl indices values are given.
    AssertionError
        If the search fails to find any of the three [uvw] rotation vectors.
    """
    # Set default box to be cubic
    if box is None:
        box = Box()
    
    # Check hkl values
    hkl = np.asarray(hkl)
    
    # Convert hkil to hkl
    if hkl.shape == (4,):
        if ishexagonal(box):
            hkl = miller.plane4to3(hkl)
            if return_hexagonal is None:
                return_hexagonal = True
        else:
            raise ValueError('Miller-Bravais indices given with non-hexagonal box')
    elif hkl.shape == (3,):
        if return_hexagonal is None:
            return_hexagonal = False
        elif return_hexagonal and not ishexagonal(box):
            raise ValueError('cannot return Miller-Bravais indices for non-hexagonal box')
    else:
        raise ValueError('Invalid hkl indices: must be 3 values or 4')
    
    if np.allclose(hkl, np.asarray(hkl, dtype=int)):
        hkl = np.asarray(hkl, dtype=int)
    else:
        raise ValueError('hkl indices must be integers')
    
    # Build conventional box if conventional_setting is given
    if conventional_setting is not None:
        primitive_box = box
        p2c_uvws = miller.vector_conventional_to_primitive(np.identity(3),
                                                           setting=conventional_setting)
        
        conventional_box = System(box=primitive_box).rotate(p2c_uvws).box
        box = conventional_box
    
    # Find two in-plane box vectors
    if hkl[0] != 0:
        if hkl[1] != 0:
            if hkl[2] != 0:
                # hkl solution
                m = np.lcm.reduce([hkl[0], hkl[1], hkl[2]]) # pylint: disable=no-member
                s = np.sign(hkl[0] * hkl[1] * hkl[2])
                a_uvw = np.array([-m / hkl[0], m / hkl[1], 0], dtype=int)
                b_uvw = np.array([-m / hkl[0], 0, m / hkl[2]], dtype=int)
            else:
                # hk0 solution
                m = np.lcm(hkl[0], hkl[1]) # pylint: disable=no-member
                s = np.sign(hkl[0] * hkl[1])
                a_uvw = np.array([-m / hkl[0], m / hkl[1], 0], dtype=int)
                b_uvw = np.array([0, 0, 1], dtype=int)
        else:
            if hkl[2] != 0:
                # h0l solution
                m = np.lcm(hkl[0], hkl[2]) # pylint: disable=no-member
                s = np.sign(hkl[0] * hkl[2])
                a_uvw = np.array([m / hkl[0], 0, -m / hkl[2]], dtype=int)
                b_uvw = np.array([0, 1, 0], dtype=int)
            else:
                # h00 solution
                m = 1
                s = np.sign(hkl[0])
                a_uvw = np.array([0, 1, 0], dtype=int)
                b_uvw = np.array([0, 0, 1], dtype=int)
    elif hkl[1] != 0:
        if hkl[2] != 0:
            # 0kl solution
            m = np.lcm(hkl[1], hkl[2]) # pylint: disable=no-member
            s = np.sign(hkl[1] * hkl[2])
            a_uvw = np.array([0, -m / hkl[1], m / hkl[2]], dtype=int)
            b_uvw = np.array([1, 0, 0], dtype=int)
        else:    
            # 0k0 solution
            m = 1
            s = np.sign(hkl[1])
            a_uvw = np.array([0, 0, 1], dtype=int)
            b_uvw = np.array([1, 0, 0], dtype=int)
    elif hkl[2] != 0:
        # 00l solution
        m = 1
        s = np.sign(hkl[2])
        a_uvw = np.array([1, 0, 0], dtype=int)
        b_uvw = np.array([0, 1, 0], dtype=int)
    else:
        raise ValueError('hkl cannot be all zeros')
    
    # Convert in-plane box vectors to primitive cell vectors if needed
    if conventional_setting is not None:
        a_uvw = miller.vector_conventional_to_primitive(a_uvw, setting=conventional_setting)
        b_uvw = miller.vector_conventional_to_primitive(b_uvw, setting=conventional_setting)
        box = primitive_box
    
    # Set default n if needed
    if maxindex is None:
        maxindex = int(np.max([np.abs(a_uvw), np.abs(b_uvw), np.abs(hkl)]))
    
    # Compute Cartesian plane normal
    planenormal = s * np.cross(np.inner(a_uvw, box.vects.T),
                               np.inner(b_uvw, box.vects.T))
    
    # Build gen_vector iterator for testing vectors
    def gen_vector(n):
        for kk in range(0, n+1):
            for sk in [1, -1]:
                k = sk * kk
                for jj in range(0, n+1):
                    for sj in [1, -1]:
                        j = sj * jj
                        for ii in range(0, n+1):
                            for si in [1, -1]:
                                i = si * ii
                                if i==0 and j==0 and k==0:
                                    continue
                                yield np.array([i, j, k], dtype=int)
    
    # First search
    a_mag = np.linalg.norm(np.inner([maxindex, maxindex, maxindex], box.vects.T))
    c_angle = 90
    a_uvw = None
    c_uvw = None
    for uvw in gen_vector(maxindex):
        cart = np.inner(uvw, box.vects.T)
        mag = np.linalg.norm(cart)
        angle = vect_angle(cart, planenormal)
        
        # Find shortest vector in the plane
        if np.isclose(np.dot(cart, planenormal), 0.0):
            if mag < a_mag:
                a_uvw = uvw
                a_mag = mag
        
        # Find vector closest to plane normal 
        elif angle < c_angle:
            c_angle = angle
            c_uvw = uvw
    
    assert a_uvw is not None, 'Failed to find first vector in slip plane'
    assert c_uvw is not None, 'Failed to find vector near slip plane normal'
    
    # Reduce c_uvw if possible
    c_uvw = c_uvw / np.gcd.reduce(np.asarray(c_uvw, dtype=int)) # pylint: disable=no-member
    
    # Second search
    a_cart = np.inner(a_uvw, box.vects.T)
    b_mag = np.linalg.norm(np.inner([maxindex, maxindex, maxindex], box.vects.T))
    b_uvw = None
    min_angle = 180.0
    for uvw in gen_vector(maxindex):
        cart = np.inner(uvw, box.vects.T)
        angle = vect_angle(a_cart, cart)
        
        # Check that vector is in plane and not parallel to a_uvw
        if np.isclose(np.dot(cart, planenormal), 0.0) and not np.isclose(angle, 0.0) and not np.isclose(angle, 180.0):
            
            # Check if right-handed
            if np.dot(np.cross(a_cart, cart), planenormal) > 0:
                
                # Find b_uvw with smallest magnitude and smallest angle
                mag = np.linalg.norm(cart)
                if (np.isclose(mag, b_mag) and angle < min_angle) or mag < b_mag:
                    b_uvw = uvw
                    b_mag = mag
                    min_angle = angle
    
    assert b_uvw is not None, 'Failed to find second vector in slip plane'
    
    # Orient the uvw sets based on cutboxvector
    if cutboxvector == 'c':
        uvws = np.array([a_uvw, b_uvw, c_uvw])
    elif cutboxvector == 'b':
        uvws = np.array([b_uvw, c_uvw, a_uvw])
    elif cutboxvector == 'a':
        uvws = np.array([c_uvw, a_uvw, b_uvw])
        
    if return_hexagonal:
        uvws = miller.vector3to4(uvws)
        
    if return_planenormal:
        return uvws, planenormal
    else:
        return uvws


# coding: utf-8
"""Provides a `duplicates_allclose` for finding duplicates in a
DataFrame with tolerances.

"""

from toolz.curried import curry, compose
import pandas as pd
import numpy as np

__all__ = ['duplicates_allclose']

pdapply = curry(pd.DataFrame.apply)  # pylint: disable=invalid-name
sort_values = curry(pd.DataFrame.sort_values)  # pylint: disable=invalid-name
pdall = curry(pd.DataFrame.all)  # pylint: disable=invalid-name
duplicated = curry(pd.DataFrame.duplicated)  # pylint: disable=invalid-name
diff = curry(pd.DataFrame.diff)  # pylint: disable=invalid-name
npappend = curry(np.append)  # pylint: disable=invalid-name


def sequence(*args):
    """Compose functions in order

    Args:
      args: the functions to compose

    Returns:
      composed functions

    >>> assert sequence(lambda x: x + 1, lambda x: x * 2)(3) == 8
    """
    return compose(*args[::-1])


@curry
def debug(statement, value):
    """Useful debug for functional programming
    """
    print()
    print(statement)
    print(value)
    return value


def find_duplicates_col(dataframe):
    """Find duplicates column by column.

    Args:
      dataframe: a dataframe

    Returns:
      a dataframe of the same shape but with bool values indicating
      neighboring duplicates in the same column

    >>> find_duplicates_col(pd.DataFrame(
    ...     dict(A=['a', 'a', 'd'],
    ...          B=['b', 'd', 'b'],
    ...          C=['d', 'c', 'c'],
    ...          D=[  1,   1,   1],
    ...          E=[  1,   2,   1])
    ... ))
           A      B      C      D      E
    0  False  False  False  False  False
    1   True  False  False   True  False
    2  False  False   True   True  False

    """
    return dataframe.apply(
        sequence(np.array, lambda x: x[:-1] == x[1:], npappend([False])), axis=0
    )


@curry
def duplicate_if_close(dataframe, fcols):
    """Find neighboring column duplicates based tolerances.

    This only works with data frames containing numbers.

    Args:
      dataframe: a dataframe
      fcols: the absolute tolerances

    Returns:
      a dataframe of the same shape but with bool values indicating
      neighboring close values in the same column

    >>> duplicate_if_close(pd.DataFrame(
    ...     dict(A=[1.01, 1.02, 1.03],
    ...          B=[1.1, 1.0, 0.99],
    ...          C=[1, 1.1, 1.2],
    ...          D=[  1,   1,   1],
    ...          E=[  1,   2,   1])
    ... ), dict(A=0.02, B=0.02, C=0.02, D=0.02, E=0.02))
           A      B      C      D      E
    0  False  False  False  False  False
    1   True  False  False   True  False
    2   True   True  False   True  False

    """
    return sequence(
        diff, pdapply(func=lambda x: np.absolute(x) <= fcols[x.name], axis=0)
    )(dataframe)


@curry
def fduplicates(dcols, fcols, dataframe):
    """Check for mixture of exact duplicates and closeness

    Args:
      dcols: the columns to check for exact duplicates
      fcols: the columns to check for closeness, a dict with column names as
       keys and tolerances as values
      dataframe: the dataframe

    Returns:
      a dataframe of the same shape but with bool values indicating
      neighboring close values in the same column

    >>> fduplicates(
    ...     dcols=['C', 'D', 'E'],
    ...     fcols=dict(A=0.02, B=0.02),
    ...     dataframe=pd.DataFrame(
    ...         dict(A=[1.01, 1.02, 1.03],
    ...              B=[1.1, 1.0, 0.99],
    ...              C=[1, 1.1, 'b'],
    ...              D=[  'a',   'a',   'a'],
    ...              E=[  1,   2,   1])
    ...     )
    ... )
           A      B      C      D      E
    0  False  False  False  False  False
    1   True  False  False   True  False
    2   True   True  False   True  False

    """
    return pd.concat(
        [
            duplicate_if_close(dataframe[list(fcols.keys())], fcols),
            find_duplicates_col(dataframe[dcols]),
        ],
        axis=1,
    )


@curry
def duplicates_allclose(dataframe, dcols, fcols):
    """
    Determine duplicates in dataframe based on tolerances.  The implementation
    first uses `pandas.DataFrame.duplicated` on the `dcols` argument with
    `keep=False` to keep all duplicates.  The duplicate sub-dataframe is then
    sorted on both `dcols` and `fcols`.  A diff between each row is then done
    on the sorted duplicates dataframe.  The float values are then checked for
    their tolerances.
    
    Note: False duplicates may be identified if tolerance ranges overlap.
    Consider dataframe with rows 1,2,3.  If row 2 matches row 1 within the
    tolerances, and row 3 matches row 2 within the tolerances, both rows 2 and
    3 will be labeled as tolerances even if row 3 does not match row 1 within
    the tolerances.
    
    Parameters
    ----------
    dataframe : pandas.DataFrame
        The dataframe to search for duplicates
    dcols : list
        The column names that are tested for exact duplicates.
    fcols : dict
        The column names (keys) that are tested using absolute tolerances
        (values).
    
    Returns
    -------
    list of bool of length nrows
        False for first occurrence of checked values, True for subsequent
        duplicates.
    """

    alltrue = lambda x: [True] * len(x)

    func = sequence(
        duplicated(subset=dcols, keep=False) if dcols else alltrue,
        lambda x: dataframe[x],
        sort_values(by=dcols + list(fcols.keys())),
        fduplicates(dcols, fcols), # pylint: disable=no-value-for-parameter
        pdapply(func=pdall, axis=1),
    )

    return (
        dataframe.assign(duplicates=func(dataframe)).duplicates.fillna(False).rename()
    )



# coding: utf-8
# Standard Python libraries
from copy import deepcopy
import warnings

# http://www.numpy.org/
import numpy as np

# atomman imports
from ..core import Atoms, Box, System
import atomman.unitconvert as uc

def dislocation_array(system, dislsol=None, m=None, n=None, burgers=None,
                      bwidth=None, cutoff=None):
    """
    Method that converts a bulk crystal system into a periodic array of
    dislocations.  A single dislocation is inserted using a dislocation
    solution.  The system's box and pbc are altered such that the system is
    periodic and compatible across the two box vectors contained in the slip
    plane.  The third box vector is non-periodic, resulting in free surfaces
    parallel to the dislocation's slip plane.
    
    Parameters
    ----------
    system : atomman.System
        A perfect, bulk atomic system.
    dislsol : atomman.defect.Stroh or atomman.defect.IsotropicVolterra, optional
        A dislocation solution to use to displace atoms by.  If not given,
        all atoms will be given linear displacements associated with the
        long-range limits.
    m : array-like object, optional
        The dislocation solution m unit vector.  This vector is in the slip
        plane and perpendicular to the dislocation line direction.  Only needed
        if dislsol is not given.
    n : array-like object, optional
        The dislocation solution n unit vector.  This vector is normal to the 
        slip plane.  Only needed if dislsol is not given.
    burgers : array-like object, optional
        The Cartesian Burger's vector for the dislocation relative to the
        given system's Cartesian coordinates.  Only needed if dislsol is not
        given.
    bwidth : float, optional
        The width of the boundary region at the free surfaces.  Atoms within
        the boundaries will be displaced by linear displacements instead of
        by the dislocation solution.  Only given if dislsol is not None.
        Default value if dislsol is given is 10 Angstroms.
    cutoff : float, optional
        Cutoff distance to use for identifying duplicate atoms to remove.
        For dislocations with an edge component, applying the displacements
        creates an extra half-plane of atoms that will have (nearly) identical
        positions with other atoms after altering the boundary conditions.
        Default cutoff value is 0.5 Angstrom.
    
    Returns
    -------
    atomman.System
        The resulting periodic array of dislocations system.  An additional
        atoms property 'old_id' will be added to map the atoms in the defect
        system back to the associated atoms in the original system.
    """
    
    # ------------------------ Parameter handling --------------------------- #
    
    # Input parameter setup for linear gradients only
    if dislsol is None:
        if m is None or n is None:
            raise ValueError('m and n are needed if no dislsol is given')
        m = np.asarray(m)
        n = np.asarray(n)
        burgers = np.asarray(burgers)
        try:
            assert np.isclose(np.linalg.norm(m), 1.0)
            assert np.isclose(np.linalg.norm(n), 1.0)
            assert np.isclose(m.dot(n), 0.0)
        except:
            raise ValueError('m and n must be perpendicular unit vectors')
        if bwidth is not None:
            raise ValueError('bwidth not allowed if dislsol is not given')

    # Input parameter setup for dislsol
    else:
        m = dislsol.m
        n = dislsol.n
        burgers = dislsol.burgers
        if bwidth is None:
            bwidth = uc.set_in_units(10, 'angstrom')
    if cutoff is None:
        cutoff = uc.set_in_units(0.5, 'angstrom')
    
    # Extract system values
    pos = system.atoms.pos
    vects = system.box.vects
    spos = system.atoms_prop(key='pos', scale=True)
    
    # -------------------- Orientation identification ----------------------- #
    
    # Find box vector alligned with u = m x n
    u = np.cross(m,n)
    line = np.isclose(np.abs(np.dot(u, vects)), np.linalg.norm(vects, axis=1))
    if np.sum(line) != 1:
        raise ValueError('box vector aligned with u = m x n not found')
    lineindex = np.where(line)[0][0]

    # Find out-of-plane box vector
    out = ~np.isclose(np.dot(n, vects), 0.0)
    if np.sum(out) > 1:
        raise ValueError('multiple box vectors have out-of-plane components')
    pnormindex = np.where(out)[0][0]

    # Set third box vector as edge direction
    motionindex = 3 - (lineindex + pnormindex)

    # Check for atoms exactly on the slip plane
    if np.isclose(spos[:, pnormindex], 0.5, rtol=0).sum() > 0:
        raise ValueError("atom positions found on slip plane: apply a coordinate shift")
    
    # ---------------------- Boundary modification -------------------------- #
    
    # Modify box vector in the motion direction by +- burgers/2
    newvects = deepcopy(vects)
    if burgers.dot(m) > 0:
        newvects[motionindex] -= burgers / 2
    else:
        newvects[motionindex] += burgers / 2
    newbox = Box(vects=newvects, origin=system.box.origin)
        
    # Make boundary condition perpendicular to slip plane non-periodic
    newpbc = [True, True, True]
    newpbc[pnormindex] = False

    # Get length of system along motionindex 
    # WHICH IS BEST!?!?!?
    length = np.abs(vects[motionindex].dot(m))
    #length = np.linalg.norm(newvects[motionindex])
    
    # -------------------- duplicate atom identification -------------------- #
    
    # Create test system to identify "duplicate" atoms
    testsystem = System(atoms=deepcopy(system.atoms), box=newbox,
                           pbc=newpbc, symbols=system.symbols)

    # Apply linear gradient shift to all atoms
    testsystem.atoms.pos += linear_displacement(pos, burgers, length, m, n)
    testsystem.atoms.old_id = range(testsystem.natoms)

    # Identify atoms at the motionindex boundary to include in the duplicate check
    spos = testsystem.atoms_prop(key='pos', scale=True)
    sburgers = np.abs(2 * burgers[motionindex] / (length))
    boundaryatoms = testsystem.atoms[  (spos[:, motionindex] < sburgers) 
                                     | (spos[:, motionindex] > 1.0 - sburgers) ]

    # Compare distances between boundary atoms to identify duplicates
    dup_atom_ids = []
    for ni, i in enumerate(boundaryatoms.old_id[:-1]):
        js = boundaryatoms.old_id[ni+1:]
        try:
            distances = np.linalg.norm(testsystem.dvect(i, js), axis=1)
            mindistance = distances.min()
        except:
            mindistance = np.linalg.norm(testsystem.dvect(i, js))
        if mindistance < cutoff:
            dup_atom_ids.append(i)
    ii = np.ones(system.natoms, dtype=bool)
    ii[dup_atom_ids] = False

    # Count found duplicate atoms
    found = system.natoms - ii.sum()

    # Count expected number of duplicates based on volume change
    expected = system.natoms - (system.natoms * newbox.volume / system.box.volume)
    if np.isclose(expected, round(expected)):
        expected = int(round(expected))
    else:
        raise ValueError('expected number of atoms to delete not an integer: check burgers vector')
        
    # Compare found versus expected number of atoms
    if found != expected:
        raise ValueError('Deleted atom mismatch: expected %i, found %i. Adjust system dimensions and/or cutoff' %(expected, found))
    
    # ---------------------- Build dislocation system ----------------------- #
    
    # Generate new system with duplicate atoms removed
    newsystem = System(atoms=system.atoms[ii], box=newbox, pbc=newpbc,
                       symbols=system.symbols)
    
    # Define old_id so atoms in newsystem can be mapped back to system
    newsystem.atoms.old_id = np.where(ii)[0]

    if dislsol is None:
        # Use only linear displacements
        disp = linear_displacement(newsystem.atoms.pos, burgers, length, m, n)
    
    else:
        # Identify boundary atoms
        miny = system.box.origin.dot(n)
        maxy = miny + vects[pnormindex].dot(n)
        if maxy < miny:
            miny, maxy = maxy, miny
        y = newsystem.atoms.pos.dot(n)
        ii = np.where((y <= miny + bwidth) | (y >= maxy - bwidth))
        
        # Use dislsol in middle and linear displacements at boundary
        disp = dislsol.displacement(newsystem.atoms.pos)
        disp[:, pnormindex] -= disp[:, pnormindex].mean()
        disp[ii] = linear_displacement(newsystem.atoms.pos[ii], burgers,
                                       length, m, n)
    
    # Displace atoms and wrap
    newsystem.atoms.pos += disp
    newsystem.wrap()
    
    return newsystem
    
def linear_displacement(pos, burgers, length, m, n):
    """
    Computes linear displacements associated with a dislocation in a system.
    
    Parameters
    ----------
    pos : array-like object
        List of Cartesian atomic positions.
    burgers : array-like object
        The Cartesian Burgers vector
    length : float
        The total length of the system along the m direction.
    m : array-like object, optional
        The dislocation solution m unit vector.  This vector is in the slip
        plane and perpendicular to the dislocation line direction.  Only needed
        if dislsol is not given.
    n : array-like object, optional
        The dislocation solution n unit vector.  This vector is normal to the 
        slip plane.  Only needed if dislsol is not given.
    """

    return np.outer(np.sign(pos.dot(n)) * (0.25 - pos.dot(m) / (2 * length)), burgers)


from copy import deepcopy
from DataModelDict import DataModelDict as DM

from .. import System
from ..tools import identifyfamily
class CrystalPrototype():
    
    def __init__(self, model):
        self.load(model=model)

    def __str__(self):
        return f'Crystal Prototype {self.id}'

    @property
    def id(self):
        return self.__id

    @property
    def key(self):
        return self.__key

    @property
    def name(self):
        return self.__name
    
    @property
    def prototype(self):
        return self.__prototype
    
    @property
    def pearson(self):
        return self.__pearson

    @property
    def strukturbericht(self):
        return self.__strukturbericht

    @property
    def sg_number(self):
        return self.__sg_number

    @property
    def sg_hm(self):
        return self.__sg_hm

    @property
    def sg_schoenflies(self):
        return self.__sg_schoenflies

    @property
    def ucell(self):
        return self.__ucell

    def load(self, model):
        
        self.__model = model = DM(model)
        
        proto = model['crystal-prototype']
        self.__key = proto['key']
        self.__id = proto['id']
        self.__name = proto['name']
        self.__prototype = proto['prototype']
        self.__pearson = proto['Pearson-symbol']
        self.__strukturbericht = proto['Strukturbericht']
        
        self.__sg_number = proto['space-group']['number']
        self.__sg_hm = proto['space-group']['Hermann-Maguin']
        self.__sg_schoenflies = proto['space-group']['Schoenflies']
        
        self.__ucell = System(model=model)

    def asmodel(self):
        
        return deepcopy(self.__model)

    def asdict(self):
        """
        Converts the structured content to a simpler dictionary.
        
        Returns
        -------
        dict
            A dictionary representation of the record's content.
        """
        params = {}
        params['key'] = self.key
        params['id'] = self.id
        params['name'] = self.name
        params['prototype'] = self.prototype
        params['pearson'] = self.pearson
        params['strukturbericht'] = self.strukturbericht
        
        params['sg_number'] = self.sg_number
        params['sg_hm'] = self.sg_hm
        params['sg_schoenflies'] = self.sg_schoenflies
        
        params['ucell'] = ucell = self.ucell
        params['crystal_family'] = identifyfamily(ucell.box)
        params['natypes'] = ucell.natypes
        
        params['a'] = ucell.box.a
        params['b'] = ucell.box.b
        params['c'] = ucell.box.c
        params['alpha'] = ucell.box.alpha
        params['beta'] = ucell.box.beta
        params['gamma'] = ucell.box.gamma
        params['natoms'] = ucell.natoms
        
        return params


# coding: utf-8
# Standard Python libraries
from io import open
from collections import OrderedDict
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# atomman imports
import atomman.unitconvert as uc
from .atoms_prop_info import atoms_prop_info
from .velocities_prop_info import velocities_prop_info
from ...lammps import style
from .. import dump_table

def dump(system, f=None, atom_style=None, units=None, natypes=None,
         potential=None, float_format='%.13f', return_info=True,
         return_pair_info=False, safecopy=False):
    """
    Write a LAMMPS-style atom data file from a System.
    
    Parameters
    ----------
    system : atomman.System 
        The system to write to the atom data file.
    f : str or file-like object, optional
        File path or file-like object to write the content to.  If not given,
        then the content is returned as a str.
    atom_style : str, optional
        The LAMMPS atom_style option associated with the data file.  If neither
        atom_style or potential is given, will set atom_style to 'atomic'.
    units : str, optional
        The LAMMPS units option associated with the data file.  If neither
        units or potential is given, will set units 'metal'.
    natypes : int, optional
        Allows the natypes value to be manually changed.  This is needed if
        natypes needs to be greater than the current number of atypes.  If
        neither natypes or potential is given, will use system.natypes.
    potential : atomman.lammps.Potential, optional
        Potential-specific values of atom_style, units, and natypes can be
        extracted from a Potential object.  If both potential and any of the
        individual values are given, the individual values will be used.
    float_format : str, optional
        c-style formatting string for floating point values.  Default value is
        '%.13f'.
    return_info : bool, optional
        Indicates if the LAMMPS command lines associated with reading in the
        file are to be returned as a str.  Default value is True.
    return_pair_info : bool, optional
        Indicates if the LAMMPS command lines associated with setting mass,
        pair_style and pair_coeff are included in the returned info.  If True,
        potential must be given and return_info must be True.  Default value is
        False.
    safecopy : bool, optional
        The LAMMPS data format requires all atoms to be inside box bounds, i.e.
        "wrapped".  If safecopy is True then a copy of the system is made to
        keep the original unwrapped.  Default value is False.
    
    Returns
    -------
    content : str
        The data file contents.  Returned if f is not given.
    read_info : str
        The LAMMPS input command lines to read the created data file in.
        Returned if return_info is True.  If return_pair_info is also True and
        potential is given, the LAMMPS input command lines for the potential
        are also included.

    Raises
    ------
    ValueError
        If return_pair_info is True and return_info is False or potential is
        not given.
    """
    # Wrap atoms and get imageflags
    if safecopy:
        system = deepcopy(system)
    imageflags = system.wrap(return_imageflags=True)
    
    # Extract potential-based parameters
    if potential is not None:
        if units is None:
            units = potential.units
        if atom_style is None:
            atom_style = potential.atom_style
        if natypes is None:
            natypes = len(potential.normalize_symbols(system.symbols))
    
    # Set default parameter values
    else:
        if units is None:
            units = 'metal'
        if atom_style is None:
            atom_style = 'atomic'
        if natypes is None:
            natypes = system.natypes

    # Generate header info
    content = '\n%i atoms\n' % system.natoms
    content += '%i atom types\n' % natypes
    
    # Write box content
    content += box_content(system, units, float_format)

    # Write atom info
    content += atoms_content(system, imageflags, atom_style, units, float_format)
    
    # Handle velocity information if included
    if 'velocity' in system.atoms_prop():
        
        # Write velocity info
        content += '\nVelocities\n\n'
        prop_info = velocities_prop_info(atom_style, units)
        
        content += dump_table(system, prop_info=prop_info, float_format=float_format)
    
    returns = []
    
    # Save to the file-like object
    if hasattr(f, 'write'):
        f.write(content)
    
    # Save to the file name
    elif f is not None:
        with open(f, 'w') as fp:
            fp.write(content)
    
    # Return as a string
    else:
        returns.append(content)
    
    # Generate LAMMPS input lines
    if return_info is True:
        read_info = info_content(system, f, atom_style=atom_style, units=units,
                                 potential=potential,
                                 return_pair_info=return_pair_info)
        returns.append(read_info)
    elif return_pair_info is True:
        raise ValueError('return_pair_info = True requires that return_info = True')
    
    if len(returns) == 1:
        return returns[0]
    
    elif len(returns) > 1:
        return tuple(returns)

def box_content(system, units, float_format):
    """
    Generates the data file lines associated with box dimensions.

    Parameters
    ----------
    system : atomman.System
    units : str
    float_format : str
    """

    # Get unit information according to the units style
    units_dict = style.unit(units)
    length_unit = units_dict['length']

    # Define line format strings
    xf2 = float_format + ' ' + float_format
    xf3 = float_format + ' ' + float_format + ' ' + float_format

    # Extract and convert box values
    xlo = uc.get_in_units(system.box.xlo, length_unit)
    xhi = uc.get_in_units(system.box.xhi, length_unit)
    ylo = uc.get_in_units(system.box.ylo, length_unit)
    yhi = uc.get_in_units(system.box.yhi, length_unit)
    zlo = uc.get_in_units(system.box.zlo, length_unit)
    zhi = uc.get_in_units(system.box.zhi, length_unit)
    xy = uc.get_in_units(system.box.xy, length_unit)
    xz = uc.get_in_units(system.box.xz, length_unit)
    yz = uc.get_in_units(system.box.yz, length_unit)
    
    # Write box values
    content = ''
    content += xf2 % (xlo, xhi) +' xlo xhi\n'
    content += xf2 % (ylo, yhi) +' ylo yhi\n'
    content += xf2 % (zlo, zhi) +' zlo zhi\n'
    if xy != 0.0 or xz != 0.0 or yz != 0.0:
        content += xf3 % (xy, xz, yz) + ' xy xz yz\n'

    return content

def atoms_content(system, imageflags, atom_style, units, float_format):
    
    content = ''
    content += f'\nAtoms # {atom_style}\n\n'
    prop_info = atoms_prop_info(atom_style, units)

    # Check if imageflags are needed
    if np.allclose(imageflags, 0):
        extra = None
    else:
        extra = OrderedDict()
        extra['imageflag_a'] = imageflags[:,0]
        extra['imageflag_b'] = imageflags[:,1]
        extra['imageflag_c'] = imageflags[:,2]
    
    content += dump_table(system, prop_info=prop_info, float_format=float_format, extra=extra)

    return content

def velocities_content(system, atom_style, units, float_format):
    
    content = ''

    if 'velocity' in system.atoms_prop():
        content += '\nVelocities\n\n'
        prop_info = velocities_prop_info(atom_style, units)
        
        content += dump_table(system, prop_info=prop_info, float_format=float_format)

    return content

def info_content(system, f, atom_style=None, units=None, potential=None,
                 return_pair_info=False):
    """
    Return appropriate units, atom_style, boundary, and read_data LAMMPS commands
    """
    if potential is not None and potential.pair_style == 'kim' and return_pair_info is True:
        return kim_info_content(system, f, potential)

    # Add comment line
    info = '# Script and atom data file prepared using atomman Python package\n\n'

    # Add units and atom_style values
    info += f'units {units}\n'
    info += f'atom_style {atom_style}\n\n'

    # Set boundary flags to p or m based on pbc values
    bflags = np.array(['m','m','m'])
    bflags[system.pbc] = 'p'
    info += f'boundary {bflags[0]} {bflags[1]} {bflags[2]}\n'
    
    # Set read_data command 
    if isinstance(f, str):
        info += f'read_data {f}\n'

    # Set pair_info
    if return_pair_info is True:
        if potential is None:
            raise ValueError('return_pair_info = True requires that potential be given')
        
        info += '\n'
        info += potential.pair_info(symbols=system.symbols,
                                    masses=system.masses)

    return info

def kim_info_content(system, f, potential):
    """
    Return appropriate units, atom_style, boundary, and read_data LAMMPS commands
    """
    # Add comment line
    info = '# Script and atom data file prepared using atomman Python package\n\n'

    # Generate pair_info
    pair_info_lines = potential.pair_info(symbols=system.symbols,
                                          masses=system.masses).split('\n')
    
    # Add kim init
    info += pair_info_lines[0] + '\n'

    # Set boundary flags to p or m based on pbc values
    bflags = np.array(['m','m','m'])
    bflags[system.pbc] = 'p'
    info += f'boundary {bflags[0]} {bflags[1]} {bflags[2]}\n'
    
    # Set read_data command 
    if isinstance(f, str):
        info += f'read_data {f}\n'

    # Add remaining pair info lines
    info += '\n'.join(pair_info_lines[1:]) + '\n'

    return info



# coding: utf-8
from .point import *
from .point import __all__ as point_all

from .dislocation_system_transform import dislocation_system_transform
from .dislocation_system_basis import dislocation_system_basis
from .differential_displacement import differential_displacement
from .DifferentialDisplacement import DifferentialDisplacement
from .disregistry import disregistry
from .slip_vector import slip_vector # pylint: disable=no-name-in-module
from .nye_tensor import nye_tensor
from .nye_tensor_p import nye_tensor_p
from .Strain import Strain # pylint: disable=no-name-in-module

from .VolterraDislocation import VolterraDislocation
from .Stroh import Stroh
from .IsotropicVolterraDislocation import IsotropicVolterraDislocation
from .solve_volterra_dislocation import solve_volterra_dislocation
from .dislocation_array import dislocation_array
from .Dislocation import Dislocation

from .free_surface_basis import free_surface_basis
from .FreeSurface import FreeSurface
from .GammaSurface import GammaSurface
from .StackingFault import StackingFault
from .pn_arctan_disregistry import pn_arctan_disregistry
from .SDVPN import SDVPN

__all__ = ['differential_displacement', 'disregistry', 'dislocation_array',
           'slip_vector', 'nye_tensor', 'nye_tensor_p', 'Stroh', 'FreeSurface',
           'StackingFault', 'Dislocation', 'Strain',
           'GammaSurface', 'free_surface_basis', 'pn_arctan_disregistry',
           'SDVPN']
__all__.extend(point_all)
__all__.sort()


# coding: utf-8

# atomman imports
from .. import load_system_model
from ...library import Database
from ...tools import screen_input

import pandas as pd

def load(key=None, method='dynamic', standing='good',
                   family=None, parent_key=None, potential=None,
                   potential_LAMMPS_id=None, potential_LAMMPS_key=None, potential_id=None, potential_key=None,
                   symbols=None, natoms=None, natypes=None, keyword=None,
                   verbose=False, database=None, localpath=None, local=False, remote=True):

    """
    Loads a potential-dependent relaxed crystal record from the library.  If
    multiple matches are found based on inputs a selection menu will appear.

    Parameters
    ----------
    key : str or list, optional
        UUID4 key(s) to search for.  Each entry has a unique random-generated
        UUID4 key.
    method : str or list or None, optional
        The relaxation method used.  Allowed values are dynamic, static and box.
        Default value is dynamic (the most rigorous relaxation method).  All
        will be loaded if set to None.
    standing : str or list or None, optional
        "good" records are the unique crystals found with the most rigorous
        relaxation, and with known prototypes over DFT structures.  "bad" are
        records filtered out, usually for being duplicates.  Default value is
        "good".  All will be loaded if set to None.
    family : str or atomman.library.CrystalPrototype or list, optional
        The crystal family associated with the relaxed crystal - either crystal
        prototype name or MP/OQMD database entry name.
    parent_key : str or list, optional
        The UUID4 key(s) assigned to the calculation that the record is based
        on.
    potential : atomman.lammps.Potential or list, optional
        A loaded LAMMPS potential object to limit the search by.
    potential_LAMMPS_id : str or list, optional
        The id for a LAMMPS implemented potential to limit the search by.
    potential_LAMMPS_key : str or list, optional
        The UUID4 for a LAMMPS implemented potential to limit the search by.
    potential_id : str or list, optional
        The id for a potential to limit the search by.
    potential_key : str or list, optional
        The UUID4 for a potential to limit the search by.
    symbols : str or list, optional
        Element symbols in the crystal to limit the search by.
    natypes : int or list, optional
        The number of unique element model symbols in the crystal to limit
        the search by.
    natoms : int or list, optional
        The number of unique atoms in the crystal's unit cell to limit the
        search by.
    keyword : str, optional
        If given, will limit the search to all records that contain the keyword
        substring.  Cannot be combined with any of the above parameters.
    database : atomman.library.Database, optional
        A pre-defined Database object to use.  If not given, will initialize
        a new Database object.  Passing in a database can save time if multiple
        calls are made for the same record type. 
    localpath : str, optional
        The local library path to use when initializing a new Database.  IF not
        given, will use the default localpath.  Ignored if database is given. 
    local : bool, optional
        Indicates if the Database object is to look for local records.  Default
        is True.  Ignored if database is given.
    remote : bool, optional
        Indicates if the Database object is to look for remote records.  Default
        is True.  Ignored if database is given.
    verbose : bool, optional
        If True, info messages will be printed during operations.  Default
        value is False.
    
    Returns
    -------
    system : atomman.System
        The system object generated from the relaxed crystal.
    """
    # Create Database object and load if needed
    if database is None:
        if local is True:
            database = Database(load='relaxed_crystal', localpath=localpath,
                                local=local, remote=remote, verbose=verbose)
        else:
            database = Database(local=local, remote=remote, verbose=verbose)
    
    # Get crystal prototype record
    crystal = database.get_relaxed_crystal(key=key, method=method, standing=standing,
                   family=family, parent_key=parent_key, potential=potential,
                   potential_LAMMPS_id=potential_LAMMPS_id, potential_LAMMPS_key=potential_LAMMPS_key,
                   potential_id=potential_id, potential_key=potential_key,
                   symbols=symbols, natoms=natoms, natypes=natypes, keyword=keyword,
                   verbose=verbose)
    
    # Retrieve unit cell information and set symbols
    return crystal.ucell


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

def normalize(system, return_transform=False):
    """
    The normalize function takes any arbitrary system and transforms it to
    be compatible with LAMMPS.  In particular, LAMMPS systems must have:
    1. Right-handed box vectors.
    2. avect = [lx, 0.0, 0.0]
    3. bvect = [xy, ly,  0.0]
    4. cvect = [xz, yz,  lz]
    5. All atoms initially inside the box dimensions.
    Note: large box tilt factors are not adjusted with this function.
    As such, the LAMMPS command 'box tilt large' may be needed.
    
    Parameters
    ----------
    system : atomman.System
        The system to normalize.
    return_transform : bool, optional
        Indicates if the transformation matrix used during the normalization
        is to be returned as well.  Default value is False
        
    Returns
    -------
    newsystem : atomman.System
        A new system that has been normalized.
    transform : np.ndarray
        The transformation matrix associated with the normalization.  Returned
        if return_transform is True.
    """
    
    # Create a copy of the system 
    system = deepcopy(system)
    
    # Swap cvector direction if box is left-handed.
    if np.dot(np.cross(system.box.avect, system.box.bvect), system.box.cvect) < 0:
        system.box_set(avect=system.box.avect,
                       bvect=system.box.bvect,
                       cvect=-system.box.cvect,
                       origin=system.box.origin + system.box.cvect)
    
    # vects represents Cartesian vectors before transformation
    vects = deepcopy(system.box.vects)
    
    # Rebuild box using a, b, c, alpha, beta, gamma
    system.box_set(a=system.box.a, b=system.box.b, c=system.box.c,
                   alpha=system.box.alpha, beta=system.box.beta, gamma=system.box.gamma,
                   scale=True)
    
    # Wrap to ensure all atoms are within the box
    system.wrap()
    
    # Update the transformation matrix
    transformation = np.linalg.lstsq(vects, system.box.vects, rcond=None)[0].T
    test1 = np.linalg.norm(transformation, axis=1)
    
    assert np.allclose(test1, np.array([1., 1., 1.])), '%f %f %f' % tuple(test1)
    assert np.isclose(transformation[0].dot(transformation[1]), 0.0)
    assert np.isclose(transformation[0].dot(transformation[2]), 0.0)
    assert np.isclose(transformation[1].dot(transformation[2]), 0.0)
    
    if return_transform:
        return system, transformation
    else:
        return system